class RestrictedTokenUnauthorized(Exception):
    pass


class AccessTokenUnauthorized(Exception):
    pass


class UserGroupLimitExceeded(Exception):
    pass


class UnhandledException(Exception):
    """
    Raised when unexpected behavior occurs.
    """


class ApplicationsNotFound(Exception):
    """
    Raised when application with the given parameters is not found in the database.
    """



"""
Connections to external services
"""
from typing import Optional

from sqlalchemy import create_engine
from sqlalchemy.orm.session import Session, sessionmaker

from .settings import (
    BROOD_DB_URI,
    BROOD_DB_URI_READ_ONLY,
    BROOD_POOL_SIZE,
    BROOD_DB_STATEMENT_TIMEOUT_MILLIS,
    BROOD_DB_POOL_RECYCLE_SECONDS,
)


def create_brood_engine(
    url: Optional[str],
    pool_size: int,
    statement_timeout: int,
    pool_recycle: int = BROOD_DB_POOL_RECYCLE_SECONDS,
):
    # Pooling: https://docs.sqlalchemy.org/en/14/core/pooling.html#sqlalchemy.pool.QueuePool
    # Statement timeout: https://stackoverflow.com/a/44936982
    return create_engine(
        url=url,
        pool_size=pool_size,
        pool_recycle=pool_recycle,
        connect_args={"options": f"-c statement_timeout={statement_timeout}"},
    )


engine = create_brood_engine(
    url=BROOD_DB_URI,
    pool_size=BROOD_POOL_SIZE,
    statement_timeout=BROOD_DB_STATEMENT_TIMEOUT_MILLIS,
    pool_recycle=BROOD_DB_POOL_RECYCLE_SECONDS,
)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def yield_db_session_from_env() -> Session:
    """
    Creates an active database session using configuration from the environment and yields it as
    per FastAPI docs:
    https://fastapi.tiangolo.com/tutorial/sql-databases/#create-a-dependency

    Behaves identically to db_session_from_env in all other respects.
    """
    session = SessionLocal()
    try:
        yield session
    finally:
        session.close()


# Read only
RO_engine = create_brood_engine(
    url=BROOD_DB_URI_READ_ONLY,
    pool_size=BROOD_POOL_SIZE,
    statement_timeout=BROOD_DB_STATEMENT_TIMEOUT_MILLIS,
    pool_recycle=BROOD_DB_POOL_RECYCLE_SECONDS,
)
RO_SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=RO_engine)


def yield_db_read_only_session() -> Session:
    """
    Yields read only database connection (created using environment variables).
    As per FastAPI docs:
    https://fastapi.tiangolo.com/tutorial/sql-databases/#create-a-dependency
    """
    session = RO_SessionLocal()
    try:
        yield session
    finally:
        session.close()



import os
from typing import Optional

import stripe  # type: ignore

RAW_ORIGIN = os.environ.get("BROOD_CORS_ALLOWED_ORIGINS")
if RAW_ORIGIN is None:
    raise ValueError(
        "BROOD_CORS_ALLOWED_ORIGINS environment variable must be set (comma-separated list of CORS allowed origins"
    )
ORIGINS = RAW_ORIGIN.split(",")

BUGOUT_URL = os.environ.get("BUGOUT_WEB_URL", "https://bugout.dev")

# Emails
BUGOUT_FROM_EMAIL = os.environ.get("BROOD_VERIFICATION_FROM_EMAIL", "info@bugout.dev")
SENDGRID_API_KEY = os.environ.get("BROOD_SENDGRID_API_KEY")

REQUIRE_EMAIL_VERIFICATION: bool = False
SEND_EMAIL_WELCOME: bool = True
TEMPLATE_ID_BUGOUT_WELCOME_EMAIL = os.environ.get(
    "SENDGRID_TEMPLATE_ID_BUGOUT_WELCOME_EMAIL"
)
TEMPLATE_ID_MOONSTREAM_WELCOME_EMAIL = os.environ.get(
    "SENDGRID_TEMPLATE_ID_MOONSTREAM_WELCOME_EMAIL"
)
MOONSTREAM_APPLICATION_ID = os.environ.get("MOONSTREAM_APPLICATION_ID")

# Database
BROOD_DB_URI = os.environ.get("BROOD_DB_URI")
if BROOD_DB_URI is None:
    raise ValueError("BROOD_DB_URI environment variable not set")
BROOD_DB_URI_READ_ONLY = os.environ.get("BROOD_DB_URI_READ_ONLY")
if BROOD_DB_URI_READ_ONLY is None:
    raise ValueError("BROOD_DB_URI_READ_ONLY environment variable not set")

BROOD_POOL_SIZE_RAW = os.environ.get("BROOD_POOL_SIZE", 0)
try:
    if BROOD_POOL_SIZE_RAW is not None:
        BROOD_POOL_SIZE = int(BROOD_POOL_SIZE_RAW)
except:
    raise Exception(f"Could not parse BROOD_POOL_SIZE as int: {BROOD_POOL_SIZE_RAW}")

BROOD_DB_STATEMENT_TIMEOUT_MILLIS_RAW = os.environ.get(
    "BROOD_DB_STATEMENT_TIMEOUT_MILLIS"
)
BROOD_DB_STATEMENT_TIMEOUT_MILLIS = 10000
try:
    if BROOD_DB_STATEMENT_TIMEOUT_MILLIS_RAW is not None:
        BROOD_DB_STATEMENT_TIMEOUT_MILLIS = int(BROOD_DB_STATEMENT_TIMEOUT_MILLIS_RAW)
except:
    raise ValueError(
        f"BROOD_DB_STATEMENT_TIMEOUT_MILLIS must be an integer: {BROOD_DB_STATEMENT_TIMEOUT_MILLIS_RAW}"
    )

BROOD_DB_POOL_RECYCLE_SECONDS_RAW = os.environ.get("BROOD_DB_POOL_RECYCLE_SECONDS")
BROOD_DB_POOL_RECYCLE_SECONDS = 1800
try:
    if BROOD_DB_POOL_RECYCLE_SECONDS_RAW is not None:
        BROOD_DB_POOL_RECYCLE_SECONDS = int(BROOD_DB_POOL_RECYCLE_SECONDS_RAW)
except:
    raise ValueError(
        f"BROOD_DB_POOL_RECYCLE_SECONDS must be an integer: {BROOD_DB_POOL_RECYCLE_SECONDS_RAW}"
    )

# Bots
BOT_INSTALLATION_TOKEN = os.environ.get("BUGOUT_BOT_INSTALLATION_TOKEN")
BOT_INSTALLATION_TOKEN_HEADER_RAW = os.environ.get(
    "BUGOUT_BOT_INSTALLATION_TOKEN_HEADER"
)
if BOT_INSTALLATION_TOKEN_HEADER_RAW is not None:
    BOT_INSTALLATION_TOKEN_HEADER = BOT_INSTALLATION_TOKEN_HEADER_RAW
else:
    raise ValueError(
        "BUGOUT_BOT_INSTALLATION_TOKEN_HEADER environment variable must be set"
    )

STRIPE_SECRET_KEY = os.environ.get("STRIPE_SECRET_KEY")
stripe.api_key = STRIPE_SECRET_KEY
STRIPE_SIGNING_SECRET = os.environ.get("STRIPE_SIGNING_SECRET")

DEFAULT_USER_GROUP_LIMIT = 15


def group_invite_link_from_env(code: str, email: Optional[str] = None) -> str:
    bugout_url_origin = BUGOUT_URL.rstrip("/")
    group_invite_link = f"{bugout_url_origin}/invites/index.html?code={code}"
    if email is not None:
        group_invite_link = f"{group_invite_link}&email={email}"
    return group_invite_link


# OpenAPI
DOCS_TARGET_PATH = "docs"
BROOD_OPENAPI_LIST = []
BROOD_OPENAPI_LIST_RAW = os.environ.get("BROOD_OPENAPI_LIST")
if BROOD_OPENAPI_LIST_RAW is not None:
    BROOD_OPENAPI_LIST = BROOD_OPENAPI_LIST_RAW.split(",")

# Web3 signature
BUGOUT_APPLICATION_ID_HEADER_RAW = os.environ.get("BUGOUT_APPLICATION_ID_HEADER")
if BUGOUT_APPLICATION_ID_HEADER_RAW is not None:
    BUGOUT_APPLICATION_ID_HEADER = BUGOUT_APPLICATION_ID_HEADER_RAW
else:
    raise ValueError("BUGOUT_APPLICATION_ID_HEADER environment variable must be set")



import logging
from typing import Any, List, Optional
from uuid import UUID

import stripe  # type: ignore
from sqlalchemy.orm.session import Session

from .data import (
    SubscriptionSessionType,
    SubscriptionPlanType,
    SubscriptionManageResponse,
)
from .models import Group, Subscription, SubscriptionPlan, KVBrood
from .settings import BUGOUT_URL

logger = logging.getLogger(__name__)


class SubscriptionInvalidParameters(ValueError):
    """
    Raised when operations are applied to a group subscription but invalid parameters are provided.
    """


class SubscriptionPlanInvalidParameters(ValueError):
    """
    Raised when operations are applied to a subscription plan but invalid parameters
    are provided with which to specify that plan.
    """


class SubscriptionPlanNotFound(Exception):
    """
    Raised when subscription plan with the given parameters is not found in the database.
    """


class GroupSubscriptionNotFound(Exception):
    """
    Raised when group subscription with the given parameters is not found in the database.
    """


class GroupSubscriptionAlreadyExists(Exception):
    """
    Raised when group subscription already attached to group.
    """


# Subscription plans


def get_subscription_plan(
    session: Session, subscription_plan_id: UUID
) -> SubscriptionPlan:
    """
    Get Bugout Subscription Plan by provided subscription_plan_id.
    """
    subscription_plan = (
        session.query(SubscriptionPlan)
        .filter(SubscriptionPlan.id == subscription_plan_id)
        .one_or_none()
    )
    if subscription_plan is None:
        raise SubscriptionPlanNotFound(
            f"Did not find subscription plan with id={subscription_plan_id}"
        )
    return subscription_plan


def get_subscription_plans(
    db_session: Session,
    plan_type: Optional[SubscriptionPlanType] = None,
    public_or_not: Optional[bool] = None,
) -> List[SubscriptionPlan]:
    """
    Extract list of subscription plans according to parameters.
    """
    query = db_session.query(SubscriptionPlan)
    if plan_type is not None:
        query = query.filter(SubscriptionPlan.plan_type == plan_type.value)
    if public_or_not is not None:
        query = query.filter(SubscriptionPlan.public == public_or_not)

    subscription_plans = query.all()
    if len(subscription_plans) < 0:
        raise SubscriptionPlanNotFound(f"Did not find any subscription plan")

    return subscription_plans


def update_subscription_plan(
    session: Session, id: UUID, **kwargs: Optional[Any]
) -> SubscriptionPlan:
    """
    Updates Subscription Plan if new values provided.
    """
    query = session.query(SubscriptionPlan).filter(SubscriptionPlan.id == id)
    for attr, value in kwargs.items():
        if value is not None:
            query.update({getattr(SubscriptionPlan, attr): value})
    session.commit()

    subscription_plan = query.first()
    return subscription_plan


def get_subscription_plan_by_units(
    db_session: Session, plan_type: SubscriptionPlanType, units_required: int
) -> SubscriptionPlan:
    """
    Return FREE plan if negative units provided.
    """
    plan_paid_type = "PRO"
    if units_required < 0:
        plan_paid_type = "FREE"

    if plan_type == SubscriptionPlanType.seats:
        kv_key = f"BUGOUT_GROUP_{plan_paid_type}_SUBSCRIPTION_PLAN"
    elif plan_type == SubscriptionPlanType.events:
        kv_key = f"BUGOUT_HUMBUG_MONTHLY_EVENTS_{plan_paid_type}_SUBSCRIPTION_PLAN"
    else:
        logger.error(f"Unsupported plan_type: {plan_type}")
        raise SubscriptionPlanNotFound("Did not find subscription plan")

    kv_variable = (
        db_session.query(KVBrood).filter(KVBrood.kv_key == kv_key).one_or_none()
    )
    if kv_variable is None:
        raise SubscriptionPlanNotFound(
            "Did not find KVBrood variable for subscription plan"
        )

    plan = (
        db_session.query(SubscriptionPlan)
        .filter(
            SubscriptionPlan.plan_type == plan_type.value,
            SubscriptionPlan.id == kv_variable.kv_value,
        )
        .one_or_none()
    )
    if plan is None:
        logger.error(
            f"Subscription plan not found, but should exist for kv_key: {kv_key}"
        )
        raise SubscriptionPlanNotFound("Did not find subscription plan")

    return plan


def get_subscription_plan_by_stripe_data(
    db_session: Session, stripe_product_id: str, stripe_price_id: str
) -> SubscriptionPlan:
    query = db_session.query(SubscriptionPlan).filter(
        SubscriptionPlan.stripe_product_id == stripe_product_id,
        SubscriptionPlan.stripe_price_id == stripe_price_id,
    )
    plan = query.one_or_none()
    if plan is None:
        raise SubscriptionPlanNotFound("Did not find subscription plan")

    return plan


# Group subscriptions


def get_group_subscription(
    session: Session,
    group_id: Optional[UUID] = None,
    plan_id: Optional[UUID] = None,
    stripe_customer_id: Optional[str] = None,
) -> Optional[Subscription]:
    """
    Get group subscription if it exists or return None.
    """
    query = session.query(Subscription)

    if group_id is not None and plan_id is not None:
        query = query.filter(
            Subscription.group_id == group_id,
            Subscription.subscription_plan_id == plan_id,
        )
    elif stripe_customer_id is not None:
        query = query.filter(Subscription.stripe_customer_id == stripe_customer_id)
    else:
        raise SubscriptionInvalidParameters(
            "In order to get group subscription, at least one of group_id "
            "with plan_id or stripe_customer_id must be specified."
        )

    subscription = query.one_or_none()
    return subscription


def get_group_subscriptions(
    session: Session, group_id: UUID, active: Optional[bool] = None
) -> List[Subscription]:
    """
    Get group subscriptions.
    """
    query = session.query(Subscription).filter(Subscription.group_id == group_id)
    if active is not None:
        query = query.filter(Subscription.active == active)
    group_subscriptions = query.all()
    return group_subscriptions


def add_group_subscription(
    session: Session,
    group_id: UUID,
    plan_id: UUID,
    units: int,
    active: bool,
    stripe_customer_id: Optional[str] = None,
    stripe_subscription_id: Optional[str] = None,
) -> Subscription:
    """
    Add group subscription plan.
    """
    subscription = Subscription(
        group_id=group_id,
        subscription_plan_id=plan_id,
        stripe_customer_id=stripe_customer_id,
        stripe_subscription_id=stripe_subscription_id,
        units=units,
        active=active,
    )

    session.add(subscription)
    session.commit()

    return subscription


def add_volume_subscription(
    db_session: Session,
    plan: SubscriptionPlan,
    group: Group,
    units_required: int,
) -> SubscriptionManageResponse:
    """
    Generated Stripe customer or extract existen and create group subscription.
    """
    subscription = get_group_subscription(db_session, group.id, plan.id)

    # Generate new Stripe user with unpaid group subscription or retrieve existing customer.
    if subscription is None:
        customer = stripe.Customer.create(
            description=group.name,
        )
        subscription = add_group_subscription(
            session=db_session,
            group_id=group.id,
            plan_id=plan.id,
            units=0,
            active=False,
            stripe_customer_id=customer.id,
        )
    else:
        customer = stripe.Customer.retrieve(subscription.stripe_customer_id)

    session_type = SubscriptionSessionType.none
    session_id = None
    session_url = None

    if subscription.stripe_subscription_id is None:
        # Handle first subscription
        st_session = stripe_checkout(customer.id, plan, units_required)
        session_type = SubscriptionSessionType.checkout
        session_id = st_session.id
        logger.info(f"Created checkout session for customer with id: {customer.id}")
    else:
        st_session = stripe_customer_portal(subscription.stripe_customer_id)
        session_type = SubscriptionSessionType.portal
        session_url = st_session.url
        logger.info(
            f"Customer with id: {subscription.stripe_customer_id} redirected to customer portal"
        )

    return SubscriptionManageResponse(
        plan_id=plan.id,
        session_type=session_type,
        session_url=session_url,
        session_id=session_id,
    )


def add_internal_subscription(
    db_session: Session,
    plan: SubscriptionPlan,
    group: Group,
    units_required: int,
) -> SubscriptionManageResponse:
    """
    Add Free or internal subscriptions to group.
    """
    add_group_subscription(
        session=db_session,
        group_id=group.id,
        plan_id=plan.id,
        units=units_required,
        active=True,
    )
    session_type = SubscriptionSessionType.humbug
    session_id = None
    session_url = None

    return SubscriptionManageResponse(
        plan_id=plan.id,
        session_type=session_type,
        session_url=session_url,
        session_id=session_id,
    )


def update_group_subscription(
    session: Session,
    group_id: UUID,
    plan_id: UUID,
    **kwargs: Optional[Any],
) -> Subscription:
    """
    Updates group subscription attributes by provided group_id and plan_id.
    """
    if plan_id is None and group_id is None:
        raise SubscriptionInvalidParameters(
            "In order to update group subscription, plan_id with group_id must be specified."
        )
    query = (
        session.query(Subscription)
        .with_for_update()
        .filter(
            Subscription.subscription_plan_id == plan_id,
            Subscription.group_id == group_id,
        )
    )
    subscription = query.one_or_none()
    if subscription is None:
        raise GroupSubscriptionNotFound("Subscription for group not found.")

    for attr, value in kwargs.items():
        if value is not None:
            query.update({getattr(Subscription, attr): value})
    session.commit()

    subscription = query.first()
    return subscription


def update_group_subscription_by_customer(
    session: Session,
    group_id: UUID,
    plan_id: UUID,
    stripe_customer_id: str,
    **kwargs: Optional[Any],
) -> Subscription:
    """
    Updates group subscription attributes by stripe_customer_id.
    """
    query = (
        session.query(Subscription)
        .with_for_update()
        .filter(
            Subscription.group_id == group_id,
            Subscription.subscription_plan_id == plan_id,
            Subscription.stripe_customer_id == stripe_customer_id,
        )
    )
    subscription = query.one_or_none()
    if subscription is None:
        raise GroupSubscriptionNotFound("Subscription for group not found.")
    for attr, value in kwargs.items():
        if value is not None:
            query.update({getattr(Subscription, attr): value})
    session.commit()

    subscription = query.first()
    return subscription


def remove_group_subscription(
    session: Session, group_id: UUID, plan_id: UUID
) -> Subscription:
    """
    Removes group subscription.
    """
    subscription = (
        session.query(Subscription)
        .with_for_update()
        .filter(
            Subscription.group_id == group_id,
            Subscription.subscription_plan_id == plan_id,
        )
        .one_or_none()
    )
    if subscription is None:
        raise GroupSubscriptionNotFound(
            f"Did not find group subscription with group id: {group_id} and plan id: {plan_id}"
        )
    session.delete(subscription)
    session.commit()

    return subscription


def stripe_customer_portal(stripe_customer_id: str) -> stripe.billing_portal.Session:
    """
    Generates portal session where user will be redirected and be able
    to manage customer, payment and subscription details.

    Docs: https://stripe.com/docs/billing/subscriptions/integrating-customer-portal#redirect
    """
    st_session = stripe.billing_portal.Session.create(
        customer=stripe_customer_id,
        return_url=BUGOUT_URL,
    )
    return st_session


def stripe_checkout(
    stripe_customer_id: str, plan: SubscriptionPlan, units_required: int
) -> stripe.checkout.Session:
    """
    Handles new subscription creation for provided stripe_customer_id according
    with Bugout Subscription Plan data.
    Param seats_required is final quantity of allowed user seats for group subscription.

    Docs: https://stripe.com/docs/api/checkout/sessions
    """
    base_url = BUGOUT_URL
    if len(base_url) > 0 and base_url[-1] == "/":
        base_url = base_url[:-1]
    redirect_url = f"{base_url}/account/teams"
    st_session = stripe.checkout.Session.create(
        customer=stripe_customer_id,
        success_url=redirect_url,
        cancel_url=redirect_url,
        payment_method_types=["card"],
        mode="subscription",
        line_items=[
            {
                "price": plan.stripe_price_id,
                "description": plan.description,
                "quantity": units_required,
            }
        ],
    )
    return st_session



"""
Brood CLI
"""
import argparse
import base64
import json
import uuid
from distutils.util import strtobool
from typing import List

from web3login.auth import to_checksum_address, verify
from web3login.exceptions import Web3VerificationError

from . import actions, data, exceptions, subscriptions
from .db import SessionLocal
from .models import (
    Application,
    Group,
    KVBrood,
    Role,
    Subscription,
    SubscriptionPlan,
    TokenType,
    User,
)


def print_user(user: User) -> None:
    """
    Print user to screen as JSON object.
    """
    print(json.dumps(actions.user_as_json_dict(user)))


def print_token(token: User) -> None:
    """
    Print token to screen as JSON object.
    """
    print(json.dumps(actions.token_as_json_dict(token)))


def print_verification_email(verification_email: User) -> None:
    """
    Print verification_email to screen as JSON object.
    """
    print(json.dumps(actions.verification_email_as_json_dict(verification_email)))


def print_group(group: Group) -> None:
    """
    Print group to screen as JSON object.
    """
    print(json.dumps(actions.group_as_json_dict(group)))


def print_group_users(group_users: data.GroupUserResponse) -> None:
    """
    Print group_users to screen as JSON object.
    """
    print(json.dumps(actions.group_users_as_json_dict(group_users)))


def users_create_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users create" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.create_user(
            session=session,
            username=args.username,
            email=args.email,
            password=args.password,
            signature=args.signature,
            is_verify=args.verified,
        )

        print_user(user)
    finally:
        session.close()


def users_update_handler(args: argparse.Namespace) -> None:
    """
    Handler for "user update" subcommand.
    """
    if args.web3_signature is None:
        raise Exception("No arguments specified to update")

    session = SessionLocal()
    try:
        query = session.query(User).filter(User.id == args.id)
        user = query.one_or_none()
        if user is None:
            raise Exception("User not found")

        if args.web3_signature is not None:
            payload_json = base64.decodebytes(args.web3_signature.encode()).decode(
                "utf-8"
            )
            payload = json.loads(payload_json)
            verified = verify(
                authorization_payload=payload,
                application_to_check=str(user.application_id)
                if user.application_id is not None
                else "",
            )
            if not verified:
                raise Web3VerificationError("Web3 registration verification error")
            web3_address = payload.get("address")
            if web3_address is None:
                raise Exception(
                    f"Web3 address in payload could not be None for user with username: {user.username}"
                )
            web3_address = to_checksum_address(web3_address)
            query.update({User.web3_address: web3_address})

        session.commit()
        print_user(user)
    except Exception as e:
        print(e)
    finally:
        session.close()


def users_get_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users get" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.get_user(session, username=args.username, email=args.email)
        print_user(user)
    finally:
        session.close()


def users_delete_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users delete" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.delete_user(session, args.username, args.email)
        print_user(user)
    finally:
        session.close()


def users_newpassword_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users newpassword" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.change_password(
            session,
            new_password=args.new_password,
            current_password=args.current_password,
            username=args.username,
            email=args.email,
        )
        print_user(user)
    finally:
        session.close()


def users_forcepassword_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users forcepassword" subcommand.
    """
    session = SessionLocal()
    password_context = actions.get_password_context()
    try:
        new_password_hash = password_context.hash(args.new_password)
        user = actions.get_user(session, args.username, args.email)
        user.password_hash = new_password_hash
        session.add(user)
        session.commit()
        print_user(user)
    finally:
        session.close()


def limits_get_group_limit_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users get_group_limit" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.get_user(session, args.username)
        group_limit = actions.get_user_group_limit(session, user.id)
        print(group_limit)
    finally:
        session.close()


def limits_set_group_limit_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users set_group_limit" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.get_user(session, args.username)
        actions.set_user_group_limit(session, user.id, args.limit)
        print(args.limit)
    finally:
        session.close()


def verifications_send_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users send" subcommand.
    """
    session = SessionLocal()
    try:
        verification_email = actions.send_verification_email(
            session, username=args.username, email=args.email
        )
        print_verification_email(verification_email)
    finally:
        session.close()


def verifications_complete_handler(args: argparse.Namespace) -> None:
    """
    Handler for "users complete" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.complete_verification(
            session, args.code, username=args.username, email=args.email
        )
        print_user(user)
    finally:
        session.close()


def tokens_create_handler(args: argparse.Namespace) -> None:
    """
    Handler for "tokens create" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.get_user(session, username=args.username, email=args.email)
        token = actions.create_token(
            session,
            user_id=user.id,
            token_type=TokenType.bugout,
            token_note="Bugout CLI token",
            restricted=args.restricted,
        )

        print_token(token)
    finally:
        session.close()


def tokens_get_handler(args: argparse.Namespace) -> None:
    """
    Handler for "tokens get" subcommand.
    """
    session = SessionLocal()
    try:
        token = actions.get_token(session, args.token)
        print_token(token)
    finally:
        session.close()


def tokens_revoke_handler(args: argparse.Namespace) -> None:
    """
    Handler for "tokens revoke" subcommand.
    """
    session = SessionLocal()
    try:
        if args.force:
            token_obj = actions.get_token(session, args.token)
            session.delete(token_obj)
            session.commit()
        else:
            token = actions.revoke_token(session, args.token)
            print_token(token)
    except:
        session.rollback()
    finally:
        session.close()


# Groups
def group_create_handler(args: argparse.Namespace) -> None:
    """
    Handler for "groups create" subcommand.
    """
    session = SessionLocal()
    try:
        user = actions.get_user(session, username=args.username)
        group = actions.create_group(session, args.name, user)
        print_group(group)
    finally:
        session.close()


def group_get_handler(args: argparse.Namespace) -> None:
    session = SessionLocal()
    try:
        if args.id is not None:
            groups_obj = actions.get_group(session, group_id=args.id)
        elif args.name is not None:
            groups_obj = actions.get_group(session, group_name=args.name)
        else:
            raise actions.GroupInvalidParameters(
                "In order to get group, at least one of group_id, or group_name must be specified"
            )
        groups = data.GroupListResponse(
            groups=[
                data.GroupResponse(
                    id=group.id,
                    name=group.name,
                    autogenerated=group.autogenerated,
                    subscriptions=[
                        subscription_plan.subscription_plan_id
                        for subscription_plan in group.subscriptions
                    ],
                    created_at=group.created_at,
                    updated_at=group.updated_at,
                )
                for group in groups_obj
            ]
        )
        print(groups.json())
    finally:
        session.close()


def group_list_handler(args: argparse.Namespace) -> None:
    session = SessionLocal()
    try:
        groups_obj = session.query(Group).all()
        groups = data.GroupListResponse(
            groups=[
                data.GroupResponse(
                    id=group.id,
                    name=group.name,
                    autogenerated=group.autogenerated,
                    subscriptions=[
                        subscription_plan.subscription_plan_id
                        for subscription_plan in group.subscriptions
                    ],
                    created_at=group.created_at,
                    updated_at=group.updated_at,
                )
                for group in groups_obj
            ]
        )
        print(groups.json())
    finally:
        session.close()


def group_role_handler(args: argparse.Namespace) -> None:
    """
    Handler for "groups role" subcommand.
    """
    session = SessionLocal()
    try:
        group = actions.get_group(session, group_id=args.group)
        group_users = actions.set_user_in_group(
            session,
            group_id=group.id,
            current_user_type=Role.owner,
            current_user_autogenerated=True,
            user_type=args.type,
            username=args.username,
        )
        print_group_users(group_users)
    finally:
        session.close()


def group_add_subscription_handler(args: argparse.Namespace) -> None:
    """
    Handler for "groups subscription add" subcommand.
    """
    session = SessionLocal()
    try:
        units = args.units
        applied_group_subscriptions = []
        plan = subscriptions.get_subscription_plan(session, args.plan)
        if plan.default_units is not None:
            units = plan.default_units

        for group_id in args.groups:
            group = session.query(Group).filter(Group.id == group_id).one_or_none()
            if group is None:
                continue
            group_subscription_ex = (
                session.query(Subscription)
                .filter(
                    Subscription.subscription_plan_id == plan.id,
                    Subscription.group_id == group.id,
                )
                .one_or_none()
            )
            if group_subscription_ex is not None:
                continue

            subscription_obj = subscriptions.add_group_subscription(
                session=session,
                group_id=group.id,
                plan_id=args.plan,
                units=units,
                active=bool(strtobool(args.active)),
                stripe_customer_id=args.stripe_customer_id,
                stripe_subscription_id=args.stripe_subscription_id,
            )
            subscription_obj_dict = actions.subscription_as_json_dict(subscription_obj)
            subscription_obj_dict["group_name"] = group.name
            applied_group_subscriptions.append(subscription_obj_dict)
        print(json.dumps(applied_group_subscriptions))
    finally:
        session.close()


def group_update_subscription_handler(args: argparse.Namespace) -> None:
    """
    Handler for "groups subscription update" subcommand.
    """
    session = SessionLocal()
    try:
        subscription_obj = subscriptions.update_group_subscription(
            session=session,
            group_id=args.group,
            plan_id=args.plan,
            stripe_customer_id=args.stripe_customer_id,
            stripe_subscription_id=args.stripe_subscription_id,
            units=args.units,
            active=bool(strtobool(args.active)) if args.active is not None else None,
        )
        print(json.dumps(actions.subscription_as_json_dict(subscription_obj)))
    finally:
        session.close()


def group_remove_subscription_handler(args: argparse.Namespace) -> None:
    """
    Handler for "groups subscription remove" subcommand.
    """
    session = SessionLocal()
    try:
        subscription_obj = subscriptions.remove_group_subscription(
            session, args.group, args.plan
        )
        print(json.dumps(actions.subscription_as_json_dict(subscription_obj)))
    finally:
        session.close()


# Subscription plans
def plan_create_handler(args: argparse.Namespace) -> None:
    """
    Creates new subscription plan.
    Controlled only by command line interface.
    """
    session = SessionLocal()
    try:
        plan_obj = SubscriptionPlan(
            name=args.name,
            description=args.description,
            stripe_product_id=args.stripe_product_id,
            stripe_price_id=args.stripe_price_id,
            default_units=args.default_units,
            plan_type=data.SubscriptionPlanType(args.plan_type).value,
            public=bool(strtobool(args.public)),
        )
        session.add(plan_obj)
        session.commit()

        if args.kv_key is not None:
            kv_brood = KVBrood(kv_key=args.kv_key, kv_value=plan_obj.id)
            session.add(kv_brood)
            session.commit()
        print(json.dumps(actions.plan_as_json_dict(plan_obj)))
    finally:
        session.close()


def plan_list_handler(args: argparse.Namespace) -> None:
    """
    Handler for "plans list" subcommand.
    """
    session = SessionLocal()
    try:
        plans_obj = subscriptions.get_subscription_plans(
            session,
        )
        plans = {
            "plans": [actions.plan_as_json_dict(plan_obj) for plan_obj in plans_obj]
        }
        print(json.dumps(plans))
    finally:
        session.close()


def plan_adopt_handler(args: argparse.Namespace) -> None:
    """
    Handler for "plans adopt" subcommand.

    This command associates the free subscription plan with all groups which do not currently have
    an associated subscription.
    """
    session = SessionLocal()
    try:
        free_plan_id = actions.get_kv_variable(
            session, "BUGOUT_GROUP_FREE_SUBSCRIPTION_PLAN"
        ).kv_value
        free_plan = subscriptions.get_subscription_plan(session, free_plan_id)

        group_ids_from_subscriptions_query = session.query(Subscription.group_id)
        orphan_group_ids_query = session.query(Group.id).filter(
            Group.id.notin_(group_ids_from_subscriptions_query)
        )
        orphan_group_ids = orphan_group_ids_query.all()
        groups_adopted = []
        for result in orphan_group_ids:
            group_id = result[0]
            session.add(
                Subscription(
                    group_id=group_id,
                    subscription_plan_id=free_plan.id,
                    seats=free_plan.default_units,
                    active=True,
                )
            )
            groups_adopted.append(str(group_id))
        session.commit()
    finally:
        session.close()

    for group_id in groups_adopted:
        print(group_id)


def plan_attr_update_handler(args: argparse.Namespace) -> None:
    """
    Handler for "plans attr update" subcommand.
    """
    session = SessionLocal()
    try:
        subscriptions.get_subscription_plan(session, args.id)

        plan_obj = subscriptions.update_subscription_plan(
            session,
            id=args.id,
            name=args.name,
            description=args.description,
            stripe_product_id=args.stripe_product_id,
            stripe_price_id=args.stripe_price_id,
            default_units=args.default_units,
            plan_type=args.plan_type,
        )
        print(json.dumps(actions.plan_as_json_dict(plan_obj)))
    finally:
        session.close()


def plan_attr_remove_handler(args: argparse.Namespace) -> None:
    """
    Removes Subscription Plan attribute value and set it as None
    if database Model allow it. Should be provided name of attribute.
    """
    session = SessionLocal()
    try:
        query = session.query(SubscriptionPlan).filter(SubscriptionPlan.id == args.id)
        query.update({getattr(SubscriptionPlan, args.attr): None})
        session.commit()

        plan_obj = query.first()
        print(json.dumps(actions.plan_as_json_dict(plan_obj)))
    finally:
        session.close()


def plan_remove_handler(args: argparse.Namespace) -> None:
    """
    Removes Bugout Subscription Plan by id.
    """
    session = SessionLocal()
    try:
        plan_obj = (
            session.query(SubscriptionPlan)
            .filter(SubscriptionPlan.id == args.id)
            .one_or_none()
        )
        if plan_obj is None:
            raise subscriptions.SubscriptionPlanNotFound(
                f"There is no Subscription Plan with id: {args.id}."
            )
        session.delete(plan_obj)
        session.commit()

        print(json.dumps(actions.plan_as_json_dict(plan_obj)))
    finally:
        session.close()


def kv_list_handler(args: argparse.Namespace) -> None:
    """
    Handler for "kv list" subcommand.
    """
    session = SessionLocal()
    try:
        kv_brood_list = session.query(KVBrood).all()
        print(
            json.dumps(
                {
                    "kv_variables": [
                        actions.kv_brood_as_json_dict(kv_brood)
                        for kv_brood in kv_brood_list
                    ]
                }
            )
        )
    finally:
        session.close()


def kv_create_handler(args: argparse.Namespace) -> None:
    """
    Handler for "kv create" subcommand.
    """
    session = SessionLocal()
    try:
        kv_brood = KVBrood(kv_key=args.kv_key, kv_value=args.kv_value)
        session.add(kv_brood)
        session.commit()

        print(json.dumps(actions.kv_brood_as_json_dict(kv_brood)))
    finally:
        session.close()


def kv_get_handler(args: argparse.Namespace) -> None:
    """
    Handler for "kv get" subcommand.
    """
    session = SessionLocal()
    try:
        kv_brood = actions.get_kv_variable(session, args.kv_key)

        print(json.dumps(actions.kv_brood_as_json_dict(kv_brood)))
    finally:
        session.close()


def kv_delete_handler(args: argparse.Namespace) -> None:
    """
    Handler for "kv delete" subcommand.
    """
    session = SessionLocal()
    try:
        kv_brood = (
            session.query(KVBrood).filter(KVBrood.kv_key == args.kv_key).one_or_none()
        )
        if kv_brood is None:
            raise actions.KVBroodNotFound("Variable not found in KVBrood table")
        session.delete(kv_brood)
        session.commit()

        print(json.dumps(actions.kv_brood_as_json_dict(kv_brood)))
    finally:
        session.close()


def applications_list_handler(args: argparse.Namespace) -> None:
    """
    Handler for "applications list" command.
    """
    session = SessionLocal()
    try:
        query = session.query(Application)
        if args.group is not None:
            query = query.filter(Application.group_id == args.group)

        applications = query.all()

        print(
            json.dumps(
                {
                    "applications": [
                        actions.application_as_json_dict(application)
                        for application in applications
                    ]
                }
            )
        )
    finally:
        session.close()


def application_create_handler(args: argparse.Namespace) -> None:
    session = SessionLocal()
    try:
        application = Application(
            group_id=args.group,
            name=args.name,
            description=args.description,
        )
        session.add(application)
        session.commit()
        print(json.dumps(actions.application_as_json_dict(application)))
    finally:
        session.close()


def application_migrate_handler(args: argparse.Namespace) -> None:
    session = SessionLocal()
    try:
        query = session.query(Application).filter(Application.id == args.application)
        application = query.one_or_none()
        if application is None:
            raise exceptions.ApplicationsNotFound("Application not found")

        query.update({Application.group_id: args.group})
        session.commit()
        print(json.dumps(actions.application_as_json_dict(application)))
    finally:
        session.close()


def main() -> None:
    parser = argparse.ArgumentParser(description="Brood CLI")
    parser.set_defaults(func=lambda _: parser.print_help())
    subcommands = parser.add_subparsers(description="Brood commands")

    parser_users = subcommands.add_parser("users", description="Brood users")
    parser_users.set_defaults(func=lambda _: parser_users.print_help())
    subcommands_users = parser_users.add_subparsers(description="Brood user commands")

    # User command parser
    parser_users_create = subcommands_users.add_parser(
        "create", description="Create Brood user"
    )
    parser_users_create.add_argument(
        "-u",
        "--username",
        required=True,
        help="Username of the user to create",
    )
    parser_users_create.add_argument(
        "-e",
        "--email",
        help="Email of the user to create",
    )
    parser_users_create.add_argument(
        "-p",
        "--password",
        help="Password of the user to create",
    )
    parser_users_create.add_argument(
        "-s", "--signature", help="Web3 signature of the user"
    )
    parser_users_create.add_argument(
        "--verified",
        action="store_true",
        help="Set this flag to create a verified user",
    )
    parser_users_create.set_defaults(func=users_create_handler)

    parser_users_update = subcommands_users.add_parser(
        "update", description="Update Brood user"
    )
    parser_users_update.add_argument(
        "-i",
        "--id",
        required=True,
        help="ID of the user to update",
    )
    parser_users_update.add_argument(
        "-w",
        "--web3_signature",
        help="Set new web3 address with provided signature",
    )
    parser_users_update.set_defaults(func=users_update_handler)

    parser_users_get = subcommands_users.add_parser("get", description="Get Brood user")
    parser_users_get.add_argument(
        "-u",
        "--username",
        help="Username of the user to get",
    )
    parser_users_get.add_argument(
        "-e",
        "--email",
        help="Email of the user to get",
    )
    parser_users_get.set_defaults(func=users_get_handler)

    parser_users_delete = subcommands_users.add_parser(
        "delete", description="Delete Brood user"
    )
    parser_users_delete.add_argument(
        "-u",
        "--username",
        help="Username of the user to delete",
    )
    parser_users_delete.add_argument(
        "-e",
        "--email",
        help="Email of the user to delete",
    )
    parser_users_delete.set_defaults(func=users_delete_handler)

    parser_users_newpassword = subcommands_users.add_parser(
        "newpassword", description="Update password"
    )
    parser_users_newpassword.add_argument(
        "-u",
        "--username",
        help="Username of the user to update password for",
    )
    parser_users_newpassword.add_argument(
        "-e",
        "--email",
        help="Email of the user to update password for",
    )
    parser_users_newpassword.add_argument(
        "-p",
        "--current-password",
        help="Current password for user",
    )
    parser_users_newpassword.add_argument("new_password", help="New password")
    parser_users_newpassword.set_defaults(func=users_newpassword_handler)

    parser_users_forcepassword = subcommands_users.add_parser(
        "forcepassword", description="Update password"
    )
    parser_users_forcepassword.add_argument(
        "-u",
        "--username",
        help="Username of the user to update password for",
    )
    parser_users_forcepassword.add_argument(
        "-e",
        "--email",
        help="Email of the user to update password for",
    )
    parser_users_forcepassword.add_argument("new_password", help="New password")
    parser_users_forcepassword.set_defaults(func=users_forcepassword_handler)

    parser_limits = subcommands.add_parser("limits", description="Brood limits")
    parser_limits.set_defaults(func=lambda _: parser_limits.print_help())
    subcommands_limits = parser_limits.add_subparsers(
        description="Brood limits commands"
    )
    parser_users_get_group_limit = subcommands_limits.add_parser(
        "get_group_limit", description="Get group limit for user"
    )
    parser_users_get_group_limit.add_argument("-u", "--username", help="Username")
    parser_users_get_group_limit.set_defaults(func=limits_get_group_limit_handler)

    parser_users_set_group_limit = subcommands_limits.add_parser(
        "set_group_limit", description="Set group limit for user"
    )
    parser_users_set_group_limit.add_argument("-u", "--username", help="Username")
    parser_users_set_group_limit.add_argument(
        "-l", "--limit", type=int, help="Group limit to assign to user"
    )
    parser_users_set_group_limit.set_defaults(func=limits_set_group_limit_handler)

    # Verification email command parser
    parser_verifications = subcommands.add_parser(
        "verifications", description="Brood user verifications"
    )
    parser_verifications.set_defaults(func=lambda _: parser_verifications.print_help())
    subcommands_verifications = parser_verifications.add_subparsers(
        description="Brood user verification commands"
    )

    parser_verifications_send = subcommands_verifications.add_parser(
        "send", description="Send Brood user verification email"
    )
    parser_verifications_send.add_argument(
        "-u",
        "--username",
        help="Username of the user to verify",
    )
    parser_verifications_send.add_argument(
        "-e",
        "--email",
        help="Email of the user to verify",
    )
    parser_verifications_send.set_defaults(func=verifications_send_handler)

    parser_verifications_complete = subcommands_verifications.add_parser(
        "complete", description="Complete Brood user verification"
    )
    parser_verifications_complete.add_argument(
        "-u",
        "--username",
        help="Username of the user to complete verification for",
    )
    parser_verifications_complete.add_argument(
        "-e",
        "--email",
        help="Email of the user to complete verification for",
    )
    parser_verifications_complete.add_argument(
        "code",
        help="Verification code to complete verification with",
    )
    parser_verifications_complete.set_defaults(func=verifications_complete_handler)

    # Tokens command parser
    parser_tokens = subcommands.add_parser("tokens", description="Brood tokens")
    parser_tokens.set_defaults(func=lambda _: parser_tokens.print_help())
    subcommands_tokens = parser_tokens.add_subparsers(
        description="Brood token commands"
    )
    parser_tokens_create = subcommands_tokens.add_parser(
        "create", description="Generate token for Brood user"
    )
    parser_tokens_create.add_argument(
        "-u",
        "--username",
        help="Username of the user to complete verification for",
    )
    parser_tokens_create.add_argument(
        "-e",
        "--email",
        help="Email of the user to complete verification for",
    )
    parser_tokens_create.add_argument(
        "-R",
        "--restricted",
        action="store_true",
        help="Set this flag to generate a restricted token for this user",
    )
    parser_tokens_create.set_defaults(func=tokens_create_handler)

    parser_tokens_get = subcommands_tokens.add_parser(
        "get", description="Get specified token"
    )
    parser_tokens_get.add_argument(
        "token", type=uuid.UUID, help="Token to retrieve from database"
    )
    parser_tokens_get.set_defaults(func=tokens_get_handler)

    parser_tokens_revoke = subcommands_tokens.add_parser(
        "revoke", description="Revoke specified token"
    )
    parser_tokens_revoke.add_argument(
        "token", type=uuid.UUID, help="Token to retrieve from database"
    )
    parser_tokens_revoke.add_argument(
        "--force",
        action="store_true",
        help="Set this flag to revoke the token no matter what",
    )
    parser_tokens_revoke.set_defaults(func=tokens_revoke_handler)

    # Groups command parser
    parser_groups = subcommands.add_parser("groups", description="Brood groups")
    parser_groups.set_defaults(func=lambda _: parser_groups.print_help())
    subcommands_groups = parser_groups.add_subparsers(
        description="Brood group commands"
    )
    parser_groups_create = subcommands_groups.add_parser(
        "create", description="Create Brood group"
    )
    parser_groups_create.add_argument(
        "-n",
        "--name",
        required=True,
        help="Group name",
    )
    parser_groups_create.add_argument(
        "-u",
        "--username",
        required=True,
        help="Username of the group owner",
    )
    parser_groups_create.set_defaults(func=group_create_handler)

    parser_groups_get = subcommands_groups.add_parser(
        "get", description="Get Brood group"
    )
    parser_groups_get.add_argument(
        "-n",
        "--name",
        help="Group name",
    )
    parser_groups_get.add_argument(
        "-i",
        "--id",
        help="Group id",
    )
    parser_groups_get.set_defaults(func=group_get_handler)

    parser_groups_list = subcommands_groups.add_parser(
        "list", description="List all Brood group"
    )
    parser_groups_list.set_defaults(func=group_list_handler)

    parser_groups_role = subcommands_groups.add_parser(
        "role", description="Add user to Brood group and set role"
    )
    parser_groups_role.add_argument(
        "-g",
        "--group",
        type=uuid.UUID,
        required=True,
        help="Group id",
    )
    parser_groups_role.add_argument(
        "-u",
        "--username",
        required=True,
        help="Added user username",
    )
    parser_groups_role.add_argument(
        "-t",
        "--type",
        required=True,
        help=f"User type: {[member for member in Role.__members__]}",
    )
    parser_groups_role.set_defaults(func=group_role_handler)

    parser_groups_subscription = subcommands_groups.add_parser(
        "subscription", description="Brood group subscription"
    )
    parser_groups_subscription.set_defaults(
        func=lambda _: parser_groups_subscription.print_help()
    )
    subcommands_groups_subscription = parser_groups_subscription.add_subparsers(
        description="Brood group subscription commands"
    )
    parser_groups_subscription_add = subcommands_groups_subscription.add_parser(
        "add", description="Add subscription plan to group"
    )
    parser_groups_subscription_add.add_argument(
        "-g",
        "--groups",
        nargs="+",
        type=uuid.UUID,
        required=True,
        help="Groups ids",
    )
    parser_groups_subscription_add.add_argument(
        "-p",
        "--plan",
        type=uuid.UUID,
        required=True,
        help="Subscription plan Bugout ID",
    )
    parser_groups_subscription_add.add_argument(
        "--stripe_customer_id",
        help="Stripe customer ID",
    )
    parser_groups_subscription_add.add_argument(
        "--stripe_subscription_id",
        help="Stripe subscription ID",
    )
    parser_groups_subscription_add.add_argument(
        "-u",
        "--units",
        help="Set units for group subscription, it will use default_units if specified for plan",
    )
    parser_groups_subscription_add.add_argument(
        "-a",
        "--active",
        choices=["True", "False"],
        help="Active status of Subscription",
    )
    parser_groups_subscription_add.set_defaults(func=group_add_subscription_handler)

    parser_groups_subscription_update = subcommands_groups_subscription.add_parser(
        "update", description="Update subscription plan for group"
    )
    parser_groups_subscription_update.add_argument(
        "-g",
        "--group",
        type=uuid.UUID,
        help="Group id",
    )
    parser_groups_subscription_update.add_argument(
        "-p",
        "--plan",
        type=uuid.UUID,
        help="Subscription plan Bugout ID",
    )
    parser_groups_subscription_update.add_argument(
        "--stripe_customer_id",
        help="Stripe customer ID",
    )
    parser_groups_subscription_update.add_argument(
        "--stripe_subscription_id",
        help="Stripe subscription ID",
    )
    parser_groups_subscription_update.add_argument(
        "-u",
        "--units",
        help="Set units for group subscription",
    )
    parser_groups_subscription_update.add_argument(
        "-a",
        "--active",
        choices=["True", "False"],
        help="Active status of Subscription",
    )
    parser_groups_subscription_update.set_defaults(
        func=group_update_subscription_handler
    )

    parser_groups_subscription_remove = subcommands_groups_subscription.add_parser(
        "remove", description="Remove subscription plan from group"
    )
    parser_groups_subscription_remove.add_argument(
        "-g",
        "--group",
        type=uuid.UUID,
        required=True,
        help="Group ID",
    )
    parser_groups_subscription_remove.add_argument(
        "-p",
        "--plan",
        type=uuid.UUID,
        required=True,
        help="Subscription plan Bugout ID",
    )
    parser_groups_subscription_remove.set_defaults(
        func=group_remove_subscription_handler
    )

    # Subscription plans command parser
    parser_plans = subcommands.add_parser(
        "plans", description="Brood subscription plans"
    )
    parser_plans.set_defaults(func=lambda _: parser_plans.print_help())
    subcommands_plans = parser_plans.add_subparsers(
        description="Brood subscription plans commands"
    )
    parser_plans_create = subcommands_plans.add_parser(
        "create", description="Create new subscription plan"
    )
    parser_plans_create.add_argument(
        "-n", "--name", required=True, help="Name of subscription plan"
    )
    parser_plans_create.add_argument(
        "-d", "--description", help="Description of subscription plan"
    )
    parser_plans_create.add_argument(
        "--stripe_product_id", help="Srtipe product ID if exists"
    )
    parser_plans_create.add_argument(
        "--stripe_price_id", help="Srtipe price ID if exists"
    )
    parser_plans_create.add_argument(
        "--kv_key",
        help="If kv_key name provided it will create a row in"
        "KVBrood table with kv_key name and kv_value as subscription plan id",
    )
    parser_plans_create.add_argument("--default_units", help="Default units for plan")
    parser_plans_create.add_argument(
        "--plan_type",
        required=True,
        help=f"Plan type: {[member for member in data.SubscriptionPlanType.__members__]}",
    )
    parser_plans_create.add_argument(
        "-p",
        "--public",
        choices=["True", "False"],
        help="Public or not Subscription Plan",
    )
    parser_plans_create.set_defaults(func=plan_create_handler)

    parser_plans_list = subcommands_plans.add_parser(
        "list", description="Get list of subscription plans"
    )
    parser_plans_list.set_defaults(func=plan_list_handler)

    parser_plans_attr = subcommands_plans.add_parser(
        "attr", description="Brood subscription plan attributes"
    )
    parser_plans_attr.set_defaults(func=lambda _: parser_plans_attr.print_help())
    subcommands_plans_attr = parser_plans_attr.add_subparsers(
        description="Brood subscription plan attribute commands"
    )
    subcommands_plans_attr_update = subcommands_plans_attr.add_parser(
        "update", description="Update subscription plan attribute"
    )
    subcommands_plans_attr_update.add_argument(
        "-i", "--id", type=uuid.UUID, required=True, help="Subscription plan Bugout ID"
    )
    subcommands_plans_attr_update.add_argument(
        "-n", "--name", help="Name of subscription plan"
    )
    subcommands_plans_attr_update.add_argument(
        "-d", "--description", help="Description of subscription plan plan"
    )
    subcommands_plans_attr_update.add_argument(
        "--stripe_product_id", help="Srtipe product ID if exists"
    )
    subcommands_plans_attr_update.add_argument(
        "--stripe_price_id", help="Srtipe price ID if exists"
    )
    subcommands_plans_attr_update.add_argument(
        "--default_units", help="Default units for plan"
    )
    subcommands_plans_attr_update.add_argument(
        "--plan_type",
        help=f"Plan type: {[member for member in data.SubscriptionPlanType.__members__]}",
    )
    subcommands_plans_attr_update.set_defaults(func=plan_attr_update_handler)
    parser_plans_attr_remove = subcommands_plans_attr.add_parser(
        "remove", description="Clear Subscription Plan attribute"
    )
    parser_plans_attr_remove.add_argument(
        "-i",
        "--id",
        type=uuid.UUID,
        required=True,
        help="Subscription plan Bugout ID",
    )
    parser_plans_attr_remove.add_argument(
        "-a",
        "--attr",
        required=True,
        help="Attribute to clear",
    )
    parser_plans_attr_remove.set_defaults(func=plan_attr_remove_handler)

    parser_plans_remove = subcommands_plans.add_parser(
        "remove", description="Remove subscription plan"
    )
    parser_plans_remove.add_argument(
        "-i", "--id", required=True, help="Subscription plan Bugout ID"
    )
    parser_plans_remove.set_defaults(func=plan_remove_handler)

    parser_plans_adopt = subcommands_plans.add_parser(
        "adopt",
        description=(
            "Subscribe every unsubscribed group to the free plan (specified by the "
            "BUGOUT_GROUPS_FREE_SUBSCRIPTION_PLAN environment variable)"
        ),
    )
    parser_plans_adopt.set_defaults(func=plan_adopt_handler)

    # KV command parser
    parser_kv = subcommands.add_parser("kv", description="Brood kv variables")
    parser_kv.set_defaults(func=lambda _: parser_kv.print_help())
    subcommands_kv = parser_kv.add_subparsers(description="Brood kv variables commands")
    parser_kv_list = subcommands_kv.add_parser(
        "list", description="List all kv brood variables"
    )
    parser_kv_list.set_defaults(func=kv_list_handler)
    parser_kv_create = subcommands_kv.add_parser(
        "create", description="Create new kv brood variable"
    )
    parser_kv_create.add_argument(
        "-k", "--kv_key", required=True, help="Name of variable"
    )
    parser_kv_create.add_argument(
        "-v", "--kv_value", required=True, help="Variable value"
    )
    parser_kv_create.set_defaults(func=kv_create_handler)
    parser_kv_get = subcommands_kv.add_parser(
        "get", description="Get kv brood variable"
    )
    parser_kv_get.add_argument("-k", "--kv_key", required=True, help="Name of variable")
    parser_kv_get.set_defaults(func=kv_get_handler)
    parser_kv_delete = subcommands_kv.add_parser(
        "delete", description="Delete kv brood variable"
    )
    parser_kv_delete.add_argument(
        "-k", "--kv_key", required=True, help="Name of variable"
    )
    parser_kv_delete.set_defaults(func=kv_delete_handler)

    # Applications command parser
    parser_applications = subcommands.add_parser(
        "applications", description="Brood applications"
    )
    parser_applications.set_defaults(func=lambda _: parser_applications.print_help())
    subcommands_applications = parser_applications.add_subparsers(
        description="Brood application commands"
    )
    parser_applications_list = subcommands_applications.add_parser(
        "list", description="List all applications"
    )
    parser_applications_list.add_argument(
        "-g", "--group", help="Filter applications by group ID"
    )
    parser_applications_list.set_defaults(func=applications_list_handler)
    parser_applications_create = subcommands_applications.add_parser(
        "create", description="Create new application"
    )
    parser_applications_create.add_argument(
        "-g", "--group", required=True, help="Filter applications by group ID"
    )
    parser_applications_create.add_argument(
        "-n", "--name", required=True, help="Name of application"
    )
    parser_applications_create.add_argument(
        "-d", "--description", help="Description of application"
    )
    parser_applications_create.set_defaults(func=application_create_handler)
    parser_applications_migrate = subcommands_applications.add_parser(
        "migrate", description="Migrate application to another group"
    )
    parser_applications_migrate.add_argument(
        "-a", "--application", required=True, help="Applications ID"
    )
    parser_applications_migrate.add_argument(
        "-g", "--group", required=True, help="Group ID migrate to"
    )
    parser_applications_migrate.set_defaults(func=application_migrate_handler)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel, Field


class ResourcePermissions(Enum):
    ADMIN = "admin"
    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    DELETE = "delete"


class HolderType(Enum):
    user = "user"
    group = "group"


class ResourceCreationRequest(BaseModel):
    application_id: UUID
    resource_data: Dict[str, Any]


class ResourceResponse(BaseModel):
    id: UUID
    application_id: UUID
    resource_data: Dict[str, Any]
    created_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class ResourcesListResponse(BaseModel):
    resources: List[ResourceResponse] = Field(default_factory=list)


class ResourceDataUpdateRequest(BaseModel):
    update: Dict[str, Any]
    drop_keys: List[str] = Field(default_factory=list)


class ResourceHolderResponse(BaseModel):
    id: UUID
    holder_type: HolderType
    permissions: Optional[List[ResourcePermissions]] = Field(default_factory=list)


class ResourceHoldersListResponse(BaseModel):
    resource_id: UUID
    holders: List[ResourceHolderResponse] = Field(default_factory=list)


class ResourcePermissionsRequest(BaseModel):
    holder_id: UUID
    holder_type: HolderType
    permissions: List[ResourcePermissions] = Field(default_factory=list)



"""
Brood resources CLI
"""
import argparse
import json

from .models import Resource
from ..db import SessionLocal


def resources_list_handler(args: argparse.Namespace) -> None:
    session = SessionLocal()
    try:
        query = session.query(Resource)
        if args.application is not None:
            query = query.filter(Resource.application_id == args.application)
        resources = query.all()

        print(
            json.dumps(
                {
                    "resources": [
                        {
                            "id": str(resource.id),
                            "application_id": str(resource.application_id),
                            "resource_data": resource.resource_data,
                            "created_at": str(resource.created_at),
                            "updated_at": str(resource.updated_at),
                        }
                        for resource in resources
                    ]
                }
            )
        )
    finally:
        session.close()


def resources_create_handler(args: argparse.Namespace) -> None:
    session = SessionLocal()
    try:
        resource = Resource(
            application_id=args.application,
            resource_data=json.loads(args.data),
        )
        session.add(resource)
        session.commit()

        print(
            json.dumps(
                {
                    "id": str(resource.id),
                    "application_id": str(resource.application_id),
                    "resource_data": resource.resource_data,
                    "created_at": str(resource.created_at),
                    "updated_at": str(resource.updated_at),
                }
            )
        )
    finally:
        session.close()


def main() -> None:
    parser = argparse.ArgumentParser(description="Brood resources CLI")
    parser.set_defaults(func=lambda _: parser.print_help())
    subcommands = parser.add_subparsers(description="Brood resources commands")

    parser_resources = subcommands.add_parser(
        "resources", description="Brood resources"
    )
    parser_resources.set_defaults(func=lambda _: parser_resources.print_help())
    subcommands_resources = parser_resources.add_subparsers(
        description="Brood user commands"
    )

    # Resources command parser
    parser_resources_list = subcommands_resources.add_parser(
        "list", description="List Brood resources"
    )
    parser_resources_list.add_argument(
        "-a",
        "--application",
        help="Application ID filter",
    )
    parser_resources_list.set_defaults(func=resources_list_handler)
    parser_resources_create = subcommands_resources.add_parser(
        "create", description="Create Brood resources"
    )
    parser_resources_create.add_argument(
        "-a",
        "--application",
        help="Application ID filter",
    )
    parser_resources_create.add_argument(
        "-d",
        "--data",
        help="Resource data, as ex: '{\"age\": 23}'",
    )
    parser_resources_create.set_defaults(func=resources_create_handler)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



from collections import defaultdict
import logging
from typing import Any, Dict, List, Optional, Set
from uuid import UUID

from sqlalchemy import or_
from sqlalchemy.orm.session import Session

from . import data
from . import exceptions
from . import models
from ..models import Application

logger = logging.getLogger(__name__)


def acl_auth(
    db_session: Session, user_id: str, user_group_id_list: List[str], resource_id: UUID
) -> Dict[data.HolderType, List[str]]:
    """
    Checks the authorization in ResourceHolderPermission model. If it represents
    a verified user or group user belongs to and generates dictionary with
    permissions for user and group. Otherwise raises a 403 error.
    """

    acl: Dict[data.HolderType, List[str]] = {
        data.HolderType.user: [],
        data.HolderType.group: [],
    }
    permissions = (
        db_session.query(
            models.ResourceHolderPermission.user_id,
            models.ResourceHolderPermission.group_id,
            models.ResourcePermission.permission,
        )
        .join(
            models.ResourcePermission,
            models.ResourcePermission.id
            == models.ResourceHolderPermission.permission_id,
        )
        .filter(models.ResourceHolderPermission.resource_id == resource_id)
        .filter(
            or_(
                models.ResourceHolderPermission.user_id == user_id,
                models.ResourceHolderPermission.group_id.in_(user_group_id_list),
            )
        )
        .all()
    )

    if not permissions:
        raise exceptions.PermissionsNotFound("No permissions for requested information")

    for permission in permissions:
        # [0] - user_id, [1] - group_id, [2] - permission name
        if permission[0] is not None:
            acl[data.HolderType.user].append(permission[2])
        if permission[1] is not None:
            acl[data.HolderType.group].append(permission[2])

    return acl


def acl_check(
    acl: Dict[data.HolderType, List[str]],
    required_scopes: Set[data.ResourcePermissions],
    check_type: Optional[data.HolderType] = None,
) -> None:
    """
    Checks if provided permissions from handler intersect with existing permissions for user/group.
    """
    if check_type is None:
        # [["read", "update"], ["update"]] -> ["read", "update"]
        permissions = {value for values in acl.values() for value in values}
    elif check_type in data.HolderType:
        permissions = {value for value in acl[check_type]}
    else:
        logger.warning("Provided wrong HolderType")
        raise exceptions.PermissionsNotFound("No permissions for requested information")

    required_scopes_values = {scope.value for scope in required_scopes}
    if not required_scopes_values.issubset(permissions):
        raise exceptions.PermissionsNotFound("No permissions for requested information")


def create_resource(
    db_session: Session,
    user_id: UUID,
    application_id: UUID,
    resource_data: Dict[str, Any],
) -> models.Resource:
    """
    Create new resource and permissions for that resource.
    Also attach current user to this permissions.
    """
    resource = models.Resource(
        application_id=application_id,
        resource_data=resource_data,
    )
    db_session.add(resource)
    db_session.commit()

    application = (
        db_session.query(Application).filter(Application.id == application_id).first()
    )

    for permission in data.ResourcePermissions:
        resource_permission = models.ResourcePermission(
            resource_id=resource.id,
            permission=permission.value,
        )
        db_session.add(resource_permission)
        db_session.commit()

        user_permission = models.ResourceHolderPermission(
            user_id=user_id,
            group_id=None,
            resource_id=resource.id,
            permission_id=resource_permission.id,
        )
        application_group_permission = models.ResourceHolderPermission(
            user_id=None,
            group_id=application.group_id,
            resource_id=resource.id,
            permission_id=resource_permission.id,
        )
        db_session.add(user_permission)
        db_session.add(application_group_permission)
        db_session.commit()

    return resource


def get_list_of_resources(
    db_session: Session,
    user_id: UUID,
    user_groups_ids: List[UUID],
    params: Dict[str, Any],
    application_id: Optional[str] = None,
) -> List[models.Resource]:
    """
    Return list of available resource to user.
    """
    query = (
        db_session.query(models.Resource)
        .join(
            models.ResourceHolderPermission,
            models.ResourceHolderPermission.resource_id == models.Resource.id,
        )
        .filter(
            or_(
                models.ResourceHolderPermission.user_id == user_id,
                models.ResourceHolderPermission.group_id.in_(user_groups_ids),
            )
        )
    )
    if application_id is not None:
        query = query.filter(models.Resource.application_id == application_id)
    for key, value in params.items():
        query = query.filter(models.Resource.resource_data[key].astext == value)

    resources = query.all()

    return resources


def get_resource(db_session: Session, resource_id: UUID) -> models.Resource:
    """
    Get resource by id or name.
    """
    query = db_session.query(models.Resource).filter(models.Resource.id == resource_id)
    resource = query.one_or_none()
    if resource is None:
        raise exceptions.ResourceNotFound("Not found requested resource")

    return resource


def update_resource_data(
    db_session: Session,
    resource_id: UUID,
    update_data: data.ResourceDataUpdateRequest,
) -> models.Resource:
    """
    Update resource data.
    """
    query = db_session.query(models.Resource).filter(models.Resource.id == resource_id)
    resource = query.one_or_none()
    if resource is None:
        raise exceptions.ResourceNotFound("Not found requested resource")

    # Update existing data
    resource_data = dict(resource.resource_data)
    for key, value in update_data.update.items():
        resource_data[key] = value
    # Remove data by key
    for drop_key in update_data.drop_keys:
        try:
            del resource_data[drop_key]
        except Exception:
            pass
    resource.resource_data = resource_data

    db_session.commit()

    return resource


def delete_resource(db_session: Session, resource_id: UUID) -> models.Resource:
    """
    Delete resource by id.
    """
    query = db_session.query(models.Resource).filter(models.Resource.id == resource_id)
    resource = query.one_or_none()
    if resource is None:
        raise exceptions.ResourceNotFound("Not found requested resource")

    db_session.delete(resource)
    db_session.commit()

    return resource


def add_holder_permissions(
    db_session: Session,
    resource_id: UUID,
    permissions_request: data.ResourcePermissionsRequest,
) -> None:
    """
    Create list of permissions for holder in resource.
    If permission already exists, this permissions will be passed.
    """
    holder_permissions_query = db_session.query(
        models.ResourceHolderPermission.permission_id
    ).filter(
        models.ResourceHolderPermission.resource_id == resource_id,
    )
    if permissions_request.holder_type == data.HolderType.user:
        holder_permissions_query = holder_permissions_query.filter(
            models.ResourceHolderPermission.user_id == permissions_request.holder_id
        )
    elif permissions_request.holder_type == data.HolderType.group:
        holder_permissions_query = holder_permissions_query.filter(
            models.ResourceHolderPermission.group_id == permissions_request.holder_id
        )
    else:
        raise Exception(f"Unexpected holder_type: {permissions_request.holder_type}")

    permissions_to_add_query = (
        db_session.query(models.ResourcePermission)
        .join(
            models.ResourceHolderPermission,
            models.ResourceHolderPermission.permission_id
            == models.ResourcePermission.id,
        )
        .filter(
            models.ResourceHolderPermission.resource_id == resource_id,
            models.ResourcePermission.permission.in_(
                [permission.value for permission in permissions_request.permissions]
            ),
            models.ResourcePermission.id.notin_(holder_permissions_query),
        )
    )

    new_permission = [
        models.ResourceHolderPermission(
            user_id=permissions_request.holder_id
            if permissions_request.holder_type == data.HolderType.user
            else None,
            group_id=permissions_request.holder_id
            if permissions_request.holder_type == data.HolderType.group
            else None,
            resource_id=resource_id,
            permission_id=permission.id,
        )
        for permission in permissions_to_add_query.all()
    ]

    db_session.add_all(new_permission)
    db_session.commit()


def get_resource_holders_permissions(
    db_session: Session, resource_id: UUID, holder_id: Optional[UUID] = None
) -> data.ResourceHoldersListResponse:
    """
    Get list of permissions for exact holder.
    """
    query = (
        db_session.query(
            models.ResourceHolderPermission.user_id,
            models.ResourceHolderPermission.group_id,
            models.ResourcePermission.permission,
        )
        .join(
            models.ResourcePermission,
            models.ResourcePermission.id
            == models.ResourceHolderPermission.permission_id,
        )
        .filter(models.ResourceHolderPermission.resource_id == resource_id)
    )
    if holder_id is not None:
        query = query.filter(
            or_(
                models.ResourceHolderPermission.user_id == holder_id,
                models.ResourceHolderPermission.group_id == holder_id,
            )
        )
    resource_holders_permissions = query.all()

    # Flat holders with their permissions.
    # (UUID_A, read), (UUID_A, update) -> (UUID_A, [read, update])
    users = list(
        filter(lambda holder: holder.user_id is not None, resource_holders_permissions)
    )
    users_sorted = sorted(users, key=lambda holder: holder[0])
    users_flat = defaultdict(list)
    for user in users_sorted:
        users_flat[user[0]].append(user[2])
    user_holders = [
        data.ResourceHolderResponse(
            id=user_id,
            holder_type=data.HolderType.user,
            permissions=[permission for permission in permissions],
        )
        for user_id, permissions in users_flat.items()
    ]

    groups = list(
        filter(lambda holder: holder.group_id is not None, resource_holders_permissions)
    )
    groups_sorted = sorted(groups, key=lambda holder: holder[1])
    groups_flat = defaultdict(list)
    for group in groups_sorted:
        groups_flat[group[1]].append(group[2])
    group_holders = [
        data.ResourceHolderResponse(
            id=group_id,
            holder_type=data.HolderType.group,
            permissions=[permission for permission in permissions],
        )
        for group_id, permissions in groups_flat.items()
    ]

    holders = user_holders + group_holders

    return data.ResourceHoldersListResponse(resource_id=resource_id, holders=holders)


def delete_resource_holder_permissions(
    db_session: Session,
    resource_id: UUID,
    permissions_request: data.ResourcePermissionsRequest,
) -> None:
    """
    Delete holder permissions if they exist.
    """
    holder_permissions_query = db_session.query(models.ResourceHolderPermission).filter(
        models.ResourceHolderPermission.resource_id == resource_id,
    )
    if permissions_request.holder_type == data.HolderType.user:
        holder_permissions_query = holder_permissions_query.filter(
            models.ResourceHolderPermission.user_id == permissions_request.holder_id
        )
    elif permissions_request.holder_type == data.HolderType.group:
        holder_permissions_query = holder_permissions_query.filter(
            models.ResourceHolderPermission.group_id == permissions_request.holder_id
        )
    else:
        raise Exception(f"Unexpected holder_type: {permissions_request.holder_type}")

    permissions_to_delete_query = holder_permissions_query.join(
        models.ResourcePermission,
        models.ResourcePermission.id == models.ResourceHolderPermission.permission_id,
    ).filter(
        models.ResourcePermission.permission.in_(
            [permission.value for permission in permissions_request.permissions]
        ),
    )

    permissions_to_delete = permissions_to_delete_query.all()
    if len(permissions_to_delete) == 0:
        raise exceptions.HolderPermissionsNotFound(
            f"There is no permissions for provided holder with id: {permissions_request.holder_id}"
        )

    for delete_permission in permissions_to_delete:
        db_session.delete(delete_permission)
    db_session.commit()



"""
User-related Brood operations
"""
import base64
import json
import logging
import re
import sys
import uuid
from datetime import datetime, timedelta
from random import randint
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, cast

from passlib.context import CryptContext
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail
from sqlalchemy import and_, func, or_
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Query
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.orm.session import Session
from web3login.auth import to_checksum_address, verify
from web3login.exceptions import Web3VerificationError

from . import data, exceptions, subscriptions
from .models import (
    Application,
    Group,
    GroupInvite,
    GroupUser,
    KVBrood,
    ResetPassword,
    Role,
    Subscription,
    SubscriptionPlan,
    Token,
    TokenType,
    User,
    UserGroupLimit,
    VerificationEmail,
)
from .settings import (
    BUGOUT_FROM_EMAIL,
    BUGOUT_URL,
    DEFAULT_USER_GROUP_LIMIT,
    MOONSTREAM_APPLICATION_ID,
    SENDGRID_API_KEY,
    TEMPLATE_ID_BUGOUT_WELCOME_EMAIL,
    TEMPLATE_ID_MOONSTREAM_WELCOME_EMAIL,
    group_invite_link_from_env,
)

logger = logging.getLogger(__name__)

SPACE_REGEX = re.compile(r"\s")


class PasswordInvalidParameters(ValueError):
    """
    Raised when provided password does not fit strength requirements.
    """

    generic_error_message = "Password is not strong enough:\n- Passwords must be at least 8 characters in length"


class UsernameInvalidParameters(ValueError):
    """
    Raised when provided username does not fit validation requirements.
    """


class UserIncorrectPassword(ValueError):
    """
    Raised when authentication attempt is made with the wrong password.
    """


class UserUnverified(Exception):
    """
    Raised when an unverified user tries to perform an action for which the user should be verified.
    """


class UserInvalidParameters(ValueError):
    """
    Raised when operations are applied to a user but invalid parameters are provided with which to
    specify that user.
    """


class UserNotFound(Exception):
    """
    Raised when user with the given parameters is not found in the database.
    """


class UserAlreadyExists(Exception):
    """
    Raised when given user name already exists in the database.
    """


class VerificationEmailNotFound(Exception):
    """
    Raised when no verification emails are found (for a given user).
    """


class VerificationIncorrectCode(ValueError):
    """
    Raised when verification is attempted with an incorrect code.
    """


class ResetPasswordNotAllowed(Exception):
    """
    Raised when requests for reset password repeats too often.
    """


class TokenNotFound(Exception):
    """
    Raised when token with the given parameters is not found in the database.
    """


class TokenNotActive(Exception):
    """
    Raised when token with the given parameters is not active.
    """


class GroupInvalidParameters(ValueError):
    """
    Raised when operations are applied to a group but invalid parameters are provided with which to
    specify that group.
    """


class GroupNotFound(Exception):
    """
    Raised when group with the given parameters is not found in the database.
    """


class GroupAlreadyExists(Exception):
    """
    Raised when given group name already exists in the database.
    """


class InviteNotAllowed(Exception):
    """
    Raised when requests for invite repeats too often.
    """


class InviteNotFound(Exception):
    """
    Raised when invite with the given parameters is not found in the database.
    """


class RoleAlreadyExists(Exception):
    """
    Raised when user already belongs to group.
    """


class RoleNotFound(Exception):
    """
    Raised when user does not belong group.
    """


class NoPermissions(Exception):
    """
    Raised when user does not have permissions to change.
    """


class NoInheritancePermission(Exception):
    """
    Raised when user violates inheritance rules.
    """


class TokenInvalidParameters(ValueError):
    """
    Raised when operations are applied to a token but invalid parameters are provided.
    """


class LackOfUserSpace(Exception):
    """
    Raised when group doesn't have free space.
    """


class KVBroodInvalidParameters(ValueError):
    """
    Raised when provided kv variable does not found in the database.
    """


class KVBroodNotFound(Exception):
    """
    Raised when required kv_brood key not found in the database.
    """


def user_as_json_dict(user: User) -> Dict[str, Any]:
    """
    Returns a representation of the given user as a JSON-serializable dictionary.
    """
    user_json = {
        "id": str(user.id),
        "username": user.username,
        "email": user.email,
        "normalized_email": user.normalized_email,
        "web3_address": user.web3_address,
        "verified": user.verified,
        "created_at": str(user.created_at),
        "updated_at": str(user.updated_at),
        "tokens": [token_as_json_dict(token) for token in user.tokens],
    }
    return user_json


def verification_email_as_json_dict(
    verification_email: VerificationEmail,
) -> Dict[str, Any]:
    """
    Returns a representation of the given verification email as a JSON-serializable dictionary.
    """
    verification_email_json = {
        "id": str(verification_email.id),
        "user_id": str(verification_email.user_id),
        "active": verification_email.active,
        "verification_code": verification_email.verification_code,
        "created_at": str(verification_email.created_at),
        "updated_at": str(verification_email.updated_at),
    }
    return verification_email_json


def token_as_json_dict(token: Token) -> Dict[str, Any]:
    """
    Returns a representation of the given token as a JSON-serializable dictionary.
    """
    token_json = {
        "id": str(token.id),
        "user_id": str(token.user_id),
        "active": token.active,
        "token_type": token.token_type.value,
        "note": token.note,
        "created_at": str(token.created_at),
        "updated_at": str(token.updated_at),
        "restricted": token.restricted,
    }
    return token_json


def group_as_json_dict(group: Group) -> Dict[str, Any]:
    """
    Returns a representation of the given group as a JSON-serializable dictionary.
    """
    group_json = {"id": str(group.id), "name": group.name}
    return group_json


def group_users_as_json_dict(group_users: data.GroupUserResponse) -> Dict[str, Any]:
    """
    Returns a representation of the given group_users as a JSON-serializable dictionary.
    """
    group_users_json = {
        "group_id": str(group_users.group_id),
        "user_id": str(group_users.user_id),
        "user_type": str(group_users.user_type),
    }
    return group_users_json


def plan_as_json_dict(plan: SubscriptionPlan) -> Dict[str, Any]:
    """
    Returns a representation of the given subscription plan as a JSON-serializable dictionary.
    """
    plan_json = {
        "id": str(plan.id),
        "name": plan.name,
        "description": plan.description,
        "stripe_product_id": plan.stripe_product_id,
        "stripe_price_id": plan.stripe_price_id,
        "default_units": plan.default_units,
        "plan_type": plan.plan_type,
        "public": plan.public,
    }
    return plan_json


def subscription_as_json_dict(subscription: Subscription) -> Dict[str, Any]:
    """
    Returns a representation of the given group subscription as a JSON-serializable dictionary.
    """
    subscription_json = {
        "group_id": str(subscription.group_id),
        "subscription_plan_id": str(subscription.subscription_plan_id),
        "stripe_customer_id": subscription.stripe_customer_id,
        "stripe_subscription_id": subscription.stripe_subscription_id,
        "units": subscription.units,
        "active": subscription.active,
    }
    return subscription_json


def kv_brood_as_json_dict(kv_brood: KVBrood) -> Dict[str, Any]:
    """
    Returns a representation of the given kv variables as a JSON-serializable dictionary.
    """
    kv_brood_json = {
        "kv_key": kv_brood.kv_key,
        "kv_value": kv_brood.kv_value,
    }
    return kv_brood_json


def application_as_json_dict(application: Application) -> Dict[str, Any]:
    application_json = {
        "id": str(application.id),
        "group_id": str(application.group_id),
        "name": application.name,
        "description": application.description,
    }
    return application_json


def get_password_context() -> CryptContext:
    """
    Generate a passlib context for Brood passwords.
    """
    password_context = CryptContext(schemes=["argon2"])
    return password_context


def generate_verification_code(
    randint_generator: Optional[Callable[[int, int], int]] = None,
) -> str:
    """
    Uses the given random integer generator to generate a string with six digits.
    This string can be used as a verification code for Brood user registration.
    """
    if randint_generator is None:
        randint_generator = randint

    verification_code_int = randint_generator(0, 10**6 - 1)
    verification_code_raw = str(verification_code_int)
    verification_code = f"{'0'*(6 - len(verification_code_raw))}{verification_code_raw}"
    return verification_code


def normalize_email(email: str) -> str:
    """
    Normalize an email to store it in the database.
    """
    segments = email.split("@")
    assert len(segments) > 1
    hostname = segments[-1]
    username_raw = "@".join(segments[:-1])
    username_no_dots = "".join(username_raw.split("."))
    assert username_no_dots != ""
    plus_segments = username_no_dots.split("+")
    assert len(plus_segments) > 0
    assert plus_segments[0] != ""
    normalized_username = plus_segments[0].lower()
    return f"{normalized_username}@{hostname}"


def verify_password_strength(password: str) -> None:
    required_len = 8
    if len(password) < required_len:
        raise PasswordInvalidParameters(
            f"Password must contain at least {required_len} characters"
        )


def verify_username(username: str) -> None:
    white_space_matches = SPACE_REGEX.search(username)
    if white_space_matches is not None:
        raise UsernameInvalidParameters(f"Username must not contain spaces")


def password_confirm(
    user: User,
    password: Optional[str] = None,
    password_hash: Optional[str] = None,
) -> bool:
    """
    Confirm password provided by user.
    """
    if password is None and password_hash is None:
        raise UserInvalidParameters(
            "In order to check user password, at least one of password, "
            "or password_hash must be specified"
        )
    password_context = get_password_context()
    if password is not None:
        if password_context.verify(password, user.password_hash):
            return True
    if password_hash is not None:
        if user.password_hash == password_hash:
            return True

    return False


def create_user(
    session: Session,
    username: str,
    email: Optional[str] = None,
    password: Optional[str] = None,
    signature: Optional[str] = None,
    autogenerated_user: bool = False,
    first_name: Optional[str] = None,
    last_name: Optional[str] = None,
    application_id: Optional[uuid.UUID] = None,
    is_verify: bool = False,
) -> User:
    """
    Creates a new user in the given database session and returns
    the created user object.

    According with autogenerated_user var it create bugout user for
    Slack/Github installation or normal user.

    Sessions are expected to be sqlalchemy Session objects:
    https://docs.sqlalchemy.org/en/13/orm/session_api.html#sqlalchemy.orm.session.Session
    """
    # Check username correctness
    assert username != ""
    username = username.lower()
    verify_username(username)

    is_creation_allowed = False

    # Regular user with password and email creation
    normalized_email = None
    password_hash = None
    if email is not None and password is not None:
        assert email != ""
        assert password != ""

        # Email should be stored as lowercase strings in the database
        normalized_email = normalize_email(email)
        verify_password_strength(password)
        password_context = get_password_context()
        password_hash = password_context.hash(password)

        is_creation_allowed = True

    # Web3 user with blockchain address creation
    web3_address = None
    if signature is not None:
        payload_json = base64.decodebytes(signature.encode()).decode("utf-8")
        payload = json.loads(payload_json)
        verified = verify(
            authorization_payload=payload,
            application_to_check=str(application_id)
            if application_id is not None
            else "",
        )
        if not verified:
            logger.info("Web3 registration verification error")
            raise Web3VerificationError()
        web3_address = payload.get("address")
        if web3_address is None:
            logger.error(
                f"Web3 address in payload could not be None for user with username: {username}"
            )
            raise Exception()
        web3_address = to_checksum_address(web3_address)

        is_creation_allowed = True

    if not is_creation_allowed:
        logger.info("Signature or email with password should be specified")
        raise UserInvalidParameters()

    user_object = User(
        username=username,
        email=email,
        normalized_email=normalized_email,
        password_hash=password_hash,
        web3_address=web3_address,
        verified=True if autogenerated_user else is_verify,
        autogenerated=True if autogenerated_user else False,
        first_name=first_name,
        last_name=last_name,
        application_id=application_id,
    )
    user_group_limit = UserGroupLimit(group_limit=DEFAULT_USER_GROUP_LIMIT)
    user_group_limit.user = user_object

    try:
        session.add(user_object)
        session.add(user_group_limit)
        session.commit()
    except IntegrityError:
        logger.info(
            f"User already exists with username: {username}, email: {email}, address: {web3_address if signature is not None else None}"
        )
        raise UserAlreadyExists()
    except Exception:
        logger.error(
            f"Unable to add user with username: {username}, email: {email}, address: {web3_address if signature is not None else None}"
        )
        raise Exception()

    return user_object


def get_user_query(
    query: Query,
    username: Optional[str] = None,
    email: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
    application_id: Optional[uuid.UUID] = None,
    web3_address: Optional[uuid.UUID] = None,
) -> Query:
    """
    Apply filters to user query depends on argument provided.
    """
    if username is None and email is None and user_id is None and web3_address is None:
        raise UserInvalidParameters(
            "In order to get user, at least one of username, email, user_id or web3_address must be specified"
        )

    query = query.filter(User.application_id == application_id)

    if username is not None:
        username = cast(str, username)
        username = username.lower()
        query = query.filter(User.username == username)

    normalized_email = None
    if email is not None:
        normalized_email = normalize_email(email)
        query = query.filter(User.normalized_email == normalized_email)

    if user_id is not None:
        query = query.filter(User.id == user_id)

    if web3_address is not None:
        query = query.filter(User.web3_address == web3_address)

    return query


def get_user(
    session: Session,
    username: Optional[str] = None,
    email: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
    application_id: Optional[uuid.UUID] = None,
    web3_address: Optional[uuid.UUID] = None,
) -> User:
    """
    Get a user by username, email, user_id or web3_address.
    If more than one of those fields is provided, will look for a user having ALL the given parameters.
    """
    if username is None and email is None and user_id is None and web3_address is None:
        raise UserInvalidParameters(
            "In order to get user, at least one of username, email, user_id, or web3_address must be specified"
        )

    query = session.query(User)
    query = get_user_query(
        query=query,
        username=username,
        email=email,
        user_id=user_id,
        application_id=application_id,
        web3_address=web3_address,
    )

    objects = query.all()
    if len(objects) == 0:
        raise UserNotFound(
            f"Did not find user with filters username: {username}, email: {email}, "
            f"id: {user_id}, application_id: {application_id} and web3_address: {web3_address}"
        )
    if len(objects) > 1:
        raise Exception(
            f"Too many results for user with filters username: {username}, email: {email}, "
            f"id: {user_id} and application_id: {application_id} and web3_address: {web3_address}"
        )

    return objects[0]


def get_user_with_groups(
    session: Session,
    username: Optional[str] = None,
    email: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
    application_id: Optional[uuid.UUID] = None,
    web3_address: Optional[uuid.UUID] = None,
) -> data.UserWithGroupsResponse:
    """
    Get a user by username, email, user_id or web3_address.
    If more than one of those fields is provided, will look for a user having ALL the given parameters.
    """
    if username is None and email is None and user_id is None and web3_address is None:
        raise UserInvalidParameters(
            "In order to get user, at least one of username, email, user_id, or web3_address must be specified"
        )

    query = session.query(User, Group, GroupUser)
    query = get_user_query(
        query=query,
        username=username,
        email=email,
        user_id=user_id,
        application_id=application_id,
        web3_address=web3_address,
    )
    query = (
        query.join(GroupUser, GroupUser.user_id == User.id, isouter=True)
        .join(Group, Group.id == GroupUser.group_id, isouter=True)
        .group_by(User, Group, GroupUser)
    )

    objects = query.all()
    if len(objects) == 0:
        raise UserNotFound(
            f"Did not find user with filters username: {username}, email: {email}, "
            f"id: {user_id}, application_id: {application_id} and web3_address: {web3_address}"
        )

    user = objects[0][0]
    groups = []
    for object in objects:
        group = object[1]
        group_user = object[2]
        # Skip if there are no groups
        if group is None:
            continue
        if group_user.user_id != user.id:
            logger.error(
                f"Unexpected group id: {group.id} fetched for user with id: {user.id}"
            )
            raise Exception("Unexpected group for user")
        groups.append(
            data.GroupUserResponse(
                group_id=group.id,
                user_id=group_user.user_id,
                user_type=group_user.user_type,
                autogenerated=group.autogenerated,
                group_name=group.name,
                parent=group.parent,
            )
        )

    return data.UserWithGroupsResponse(
        user_id=user.id,
        username=user.username,
        first_name=user.first_name,
        last_name=user.last_name,
        email=user.email,
        normalized_email=user.normalized_email,
        verified=user.verified,
        created_at=user.created_at,
        updated_at=user.updated_at,
        autogenerated=user.autogenerated,
        application_id=user.application_id,
        web3_address=user.web3_address,
        groups=groups,
    )


def get_current_user_by_token(session: Session, token: uuid.UUID) -> Tuple[bool, User]:
    """
    Get user by its active token for authentication middleware.
    """
    query = (
        session.query(Token.active, User)
        .join(User, User.id == Token.user_id)
        .filter(Token.id == token)
    )
    objects = query.all()
    if len(objects) == 0:
        raise TokenNotFound("Token not found")
    if len(objects) > 1:
        logger.error(f"Found several users with token id: {token}")
        raise Exception("Found several users")

    is_token_active = objects[0][0]
    user = objects[0][1]

    return is_token_active, user


def get_current_user_with_groups_by_token(
    session: Session, token: uuid.UUID
) -> Tuple[bool, data.UserWithGroupsResponse]:
    """
    Extend user data with groups for authentication middleware.
    """
    query = (
        session.query(Token.active, User, Group, GroupUser)
        .join(User, User.id == Token.user_id)
        .join(GroupUser, GroupUser.user_id == User.id, isouter=True)
        .join(Group, Group.id == GroupUser.group_id, isouter=True)
        .filter(Token.id == token)
        .group_by(Token.active, User, Group, GroupUser)
    )
    objects = query.all()
    if len(objects) == 0:
        raise TokenNotFound("Token not found")

    active_token = objects[0][0]
    user = objects[0][1]
    groups = []
    for object in objects:
        if object[1].id != user.id:
            logger.error(f"Found several users with token id: {token}")
            raise Exception("Found several users")
        # Skip if there are no groups
        if object[2] is None:
            continue
        if object[3].user_id != user.id:
            logger.error(
                f"Unexpected group id: {object[2].id} fetched for user with id: {user.id}"
            )
            raise Exception("Unexpected group in list")
        groups.append(
            data.GroupUserResponse(
                group_id=object[2].id,
                user_id=object[3].user_id,
                user_type=object[3].user_type,
                autogenerated=object[2].autogenerated,
                group_name=object[2].name,
                parent=object[2].parent,
            )
        )

    return active_token, data.UserWithGroupsResponse(
        user_id=user.id,
        username=user.username,
        first_name=user.first_name,
        last_name=user.last_name,
        email=user.email,
        normalized_email=user.normalized_email,
        verified=user.verified,
        created_at=user.created_at,
        updated_at=user.updated_at,
        autogenerated=user.autogenerated,
        application_id=user.application_id,
        groups=groups,
    )


def update_user(
    session: Session,
    user_id: uuid.UUID,
    first_name: Optional[str] = None,
    last_name: Optional[str] = None,
) -> User:
    """
    Update user's first_name and last_name with the given ID.
    """
    if first_name is None and last_name is None:
        raise UserInvalidParameters(
            "In order to update user, at least one of first_name, or last_name must be specified"
        )

    query = session.query(User).filter(User.id == user_id)
    user_object = query.first()

    if first_name is not None:
        query.update({User.first_name: first_name})
    if last_name is not None:
        query.update({User.last_name: last_name})

    session.commit()
    return user_object


def delete_user(
    session: Session,
    username: Optional[str] = None,
    email: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
    current_password: Optional[str] = None,
) -> User:
    """
    Delete a user by username, email, or user_id. Uses get_user to locate the specified user.
    """
    if username is None and email is None and user_id is None:
        raise UserInvalidParameters(
            "In order to delete user, at least one of username, email, or user_id must be specified"
        )
    user = get_user(session, username, email, user_id)

    if user.autogenerated is False:
        if current_password is None:
            raise UserIncorrectPassword(
                "Attempted to delete user please provide password"
            )
        else:
            password_abide = password_confirm(user, password=current_password)
            if password_abide is False:
                raise UserIncorrectPassword(
                    "Attempted to delete user with incorrect password"
                )

    session.delete(user)
    session.commit()
    return user


def change_password(
    session: Session,
    new_password: str,
    current_password: Optional[str] = None,
    current_password_hash: Optional[str] = None,
    username: Optional[str] = None,
    email: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
) -> User:
    """
    Change a user's password.
    """
    user = get_user(session, username, email, user_id)

    password_abide = password_confirm(
        user,
        password=current_password,
        password_hash=current_password_hash,
    )
    if password_abide is False:
        raise UserIncorrectPassword(
            "Attempted to change password with incorrect password"
        )

    verify_password_strength(new_password)

    password_context = get_password_context()
    new_password_hash = password_context.hash(new_password)
    user.password_hash = new_password_hash
    session.add(user)
    session.commit()
    return user


def send_welcome_email(user: User, application_id: Optional[uuid.UUID] = None) -> None:
    """
    Send welcome email to each new user.
    """
    user_id = str(user.id)
    if SENDGRID_API_KEY is None:
        logger.error(
            f"Missed SENDGRID_API_KEY, message was not sent to user with id: {user_id}"
        )
        return

    message = Mail(
        from_email=f"Sophia from Bugout <{BUGOUT_FROM_EMAIL}>", to_emails=user.email
    )
    message.dynamic_template_data = {"username": user.username}

    if application_id is not None:
        if str(application_id) != MOONSTREAM_APPLICATION_ID:
            logger.error(
                f"Unhandled welcome email for application with id: {str(application_id)}"
            )
            return
        else:
            message.template_id = TEMPLATE_ID_MOONSTREAM_WELCOME_EMAIL
    else:
        message.template_id = TEMPLATE_ID_BUGOUT_WELCOME_EMAIL

    logger.info(f"Sending welcome email for user with id={user_id}...")

    try:
        sg = SendGridAPIClient(SENDGRID_API_KEY)
        sg.send(message)
        logger.info(
            f"Welcome email successfully submitted to Sendgrid for user with id={user_id}"
        )
    except Exception:
        logger.error(f"Error sending welcome email to {user.email}")
        pass


def send_verification_email(
    session: Session,
    username: Optional[str] = None,
    email: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
) -> VerificationEmail:
    """
    Send verification email for the given (user identified using get_user).

    Pulls configuration from the following two environment variables:
    - BROOD_VERIFICATION_FROM_EMAIL - Email address from which verification emails are sent.
    - BROOD_SENDGRID_API_KEY - SendGrid API key
    """
    user = get_user(session, username, email, user_id)
    query = (
        session.query(VerificationEmail)
        .filter(VerificationEmail.active == True)
        .filter(VerificationEmail.user_id == user.id)
    )
    for old_email in query:
        old_email.active = False
        session.add(old_email)

    verification_code = generate_verification_code()
    verification_email = VerificationEmail(
        verification_code=verification_code,
        active=True,
        user_id=user.id,
    )
    session.add(verification_email)
    session.commit()

    if SENDGRID_API_KEY:
        message = Mail(
            from_email=BUGOUT_FROM_EMAIL,
            to_emails=user.email,
            subject="Bugout.dev account verification",
            html_content=f"Verification code: <strong>{verification_email.verification_code}</strong>",
        )

        try:
            sg = SendGridAPIClient(SENDGRID_API_KEY)
            response = sg.send(message)
        except Exception as e:
            logger.exception(e)
            raise

    return verification_email


def complete_verification(
    session: Session,
    code: str,
    username: Optional[str] = None,
    email: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
) -> User:
    """
    Complete verification flow for the given user.
    """
    user = get_user(session, username, email, user_id)
    query = (
        session.query(VerificationEmail)
        .filter(VerificationEmail.active == True)
        .filter(VerificationEmail.user_id == user.id)
    )
    verification_email = query.first()
    if verification_email is None:
        raise VerificationEmailNotFound(
            f"No verification emails found for user with ID: {user.id}"
        )

    if code != verification_email.verification_code:
        raise VerificationIncorrectCode(
            f"Code ({code}) does not match code of any active verification for user with ID {user.id}"
        )

    verification_email.active = False
    session.add(verification_email)
    user.verified = True
    session.add(user)
    session.commit()
    return user


def generate_reset_password(session: Session, email: str) -> ResetPassword:
    """
    Checks if user exists and password reset requests happens not too often.

    We use delay 1 minute to prevent spam.

    TODO(kompotkot): This function looks monstrous because of problem with UTC timezone.
    Datetime from PC: 020-10-14 17:39:08.798535+00:00
    Datetime from DB: 2020-10-14 17:49:05.567316-07:00
    It requires changes and DB migration to has possibility compare datetime outside of Postgres.
    """
    if email is None:
        raise UserInvalidParameters("In order to get user, email must be specified")

    user = get_user(session, email=email)

    query = session.query(ResetPassword).filter(ResetPassword.user_id == user.id)
    reset_object = query.one_or_none()

    if reset_object is None:
        reset_object = ResetPassword(user_id=user.id)
        session.add(reset_object)
        session.commit()

        return reset_object

    current_time = datetime.utcnow()
    ts_delay = current_time - timedelta(minutes=1)

    reset_object_fresh = query.filter(ResetPassword.created_at > ts_delay).one_or_none()

    if reset_object_fresh is not None:
        raise ResetPasswordNotAllowed(
            f"Please wait {ts_delay} until next password reset"
        )

    session.delete(reset_object)
    session.commit()

    reset_object = ResetPassword(user_id=user.id)
    session.add(reset_object)
    session.commit()

    return reset_object


def send_reset_password_email(reset_object: ResetPassword, email: str) -> None:
    """
    Send reset password for given email.
    """
    message = Mail(
        from_email=BUGOUT_FROM_EMAIL,
        to_emails=email,
        subject="Bugout.dev password reset",
        html_content=f"Reset password url: {BUGOUT_URL}/password/reset/index.html?reset_id={str(reset_object.id)}",
    )

    try:
        sg = SendGridAPIClient(SENDGRID_API_KEY)
        response = sg.send(message)
    except Exception as e:
        logger.exception(e)
        raise


def reset_password_confirmation(
    session: Session, reset_id: uuid.UUID, new_password: str
) -> User:
    """
    Process password change for user. Last step in password reset workflow.
    Mark as completed password reset in DB.
    """
    query = (
        session.query(ResetPassword)
        .filter(ResetPassword.id == reset_id)
        .filter(ResetPassword.completed == False)
    )
    reset_object = query.one_or_none()

    if reset_object is None:
        raise UserNotFound(f"Did not find user to reset password")

    verify_password_strength(new_password)

    user = get_user(session, user_id=reset_object.user_id)
    try:
        change_password(
            session,
            new_password=new_password,
            current_password_hash=user.password_hash,
            user_id=reset_object.user_id,
        )

        query.update({ResetPassword.completed: True})
        session.commit()
    except Exception as e:
        logger.error(f"Unable to change password, error: {e}")
        raise

    return user


def create_token(
    session: Session,
    user_id: uuid.UUID,
    token_type: Optional[TokenType] = TokenType.bugout,
    token_note: Optional[str] = None,
    restricted: bool = False,
) -> Token:
    """
    Generate an access token for the given user (user retrieved using get_user).
    """
    token = Token(
        user_id=user_id,
        active=True,
        token_type=token_type,
        note=token_note,
        restricted=restricted,
    )
    session.add(token)
    session.commit()
    return token


def get_token(session: Session, token: uuid.UUID) -> Token:
    """
    Retrieve the token with the given ID from the database (if it exists).
    """
    token_object = session.query(Token).filter(Token.id == token).first()
    if token_object is None:
        raise TokenNotFound(f"Token not found")
    token_object = cast(Token, token_object)
    return token_object


def get_tokens(session: Session, user_id: uuid.UUID) -> List[Token]:
    """
    Retrieve the list of tokens for user.
    """
    objects = session.query(Token).filter(Token.user_id == user_id).all()
    return objects


def update_token(
    session: Session,
    token: uuid.UUID,
    token_type: Optional[TokenType] = None,
    token_note: Optional[str] = None,
    target_token: Optional[uuid.UUID] = None,
) -> Token:
    """
    Update the token's type and note with the given ID.
    """
    if token_type is None and token_note is None:
        raise TokenInvalidParameters(
            "In order to update token, at least one of token_type, or token_note must be specified"
        )
    if token_type is not None and token_type not in TokenType:
        raise TokenInvalidParameters("Incorrect token type provided.")

    token_object = get_token(session, token)
    if token_object.restricted:
        raise exceptions.RestrictedTokenUnauthorized(
            "Restricted tokens cannot update tokens"
        )

    target_object = token_object
    if target_token is not None:
        target_object = get_token(session, target_token)
    if token_object.user_id != target_object.user_id:
        raise exceptions.AccessTokenUnauthorized(
            "Could not perform the desired operation."
        )

    if token_type is not None:
        token_object.token_type = token_type
    if token_note is not None:
        token_object.note = token_note

    session.add(token_object)
    session.commit()
    return token_object


def revoke_token(
    session: Session, token: uuid.UUID, target: Optional[uuid.UUID] = None
) -> Token:
    """
    Revoke the token with the given ID (if it exists).
    """
    token_object = get_token(session, token)
    if token_object.restricted:
        raise exceptions.RestrictedTokenUnauthorized(
            "Restricted tokens are not authorized to revoke tokens"
        )
    target_object = token_object
    if target is not None:
        target_object = get_token(session, target)
    if token_object.user_id != target_object.user_id:
        raise exceptions.AccessTokenUnauthorized(
            "Could not perform the desired operation."
        )
    target_object.active = False
    session.add(target_object)
    session.commit()
    return target_object


def login(
    session: Session,
    username: str,
    password: str,
    token_type: Optional[TokenType] = TokenType.bugout,
    token_note: Optional[str] = None,
    restricted: bool = False,
    application_id: Optional[uuid.UUID] = None,
) -> Token:
    """
    Login with the given username and password to get a new token for the user with that username.
    If token_type and token_note provieded it works as token generation handler. By default it
    creates "bugout" token with None in note.
    """
    user = get_user(session, username=username, application_id=application_id)

    password_abide = password_confirm(user, password=password)
    if password_abide is False:
        raise UserIncorrectPassword("Attempted to login with incorrect password")

    token = create_token(
        session,
        user_id=user.id,
        token_type=token_type,
        token_note=token_note,
        restricted=restricted,
    )
    return token


def get_user_limit(session: Session, group: Group, modifier: int) -> bool:
    """
    Comparing number of free seats and number of users in group and
    subgroups with parent groups.

    :return: False if space is exhausted.
    """
    max_seats = get_num_seats(session, group)
    num_users = get_num_users(session, group)

    free_space = max_seats - num_users + 1
    if modifier > 0 and free_space - modifier <= 0:
        return False

    return True


def get_group_children(
    session: Session,
    group_id: uuid.UUID,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[Group]:
    """
    Return list of children who have provided group id in parent field.
    """
    query = session.query(Group).filter(Group.parent == group_id)
    if limit is not None:
        query = query.limit(limit)
    if offset is not None:
        query = query.offset(offset)
    children = query.all()
    return children


def get_num_users(session: Session, group: Group) -> int:
    """
    Calculates number of unique users in group, group's children and
    parent if it is exists.
    """
    top_group_id = group.id
    if group.parent is not None:
        top_group_id = group.parent

    num_users = (
        session.query(func.count(func.distinct(User.id)))
        .join(GroupUser, GroupUser.user_id == User.id)
        .filter(User.autogenerated == False, GroupUser.group_id == top_group_id)
    ).one()[0]

    return num_users


def get_num_seats(session: Session, group: Group) -> int:
    """
    Calculate number of maximum seats for provided group.
    It includes all active subscriptions.
    """
    seats_plans = (
        session.query(SubscriptionPlan)
        .filter(SubscriptionPlan.plan_type == data.SubscriptionPlanType.seats.value)
        .all()
    )
    seats_plans_ids = [plan.id for plan in seats_plans]
    subscriptions = (
        session.query(Subscription)
        .filter(Subscription.active == True)
        .filter(
            or_(
                Subscription.group_id == group.parent, Subscription.group_id == group.id
            )
        )
        .all()
    )
    num_seats = sum(
        [
            subscription.units
            for subscription in subscriptions
            if subscription.subscription_plan_id in seats_plans_ids
        ]
    )
    return num_seats


def find_group(session: Session, group_id: uuid.UUID) -> Group:
    group = session.query(Group).filter(Group.id == group_id).one_or_none()
    if group is None:
        raise GroupNotFound("Did not find group")

    return group


def get_group(
    session: Session,
    group_id: Optional[uuid.UUID] = None,
    group_name: Optional[str] = None,
    user_id: Optional[uuid.UUID] = None,
) -> Group:
    """
    Get a group by group_name, or group_id. If more than one of those fields is provided, will
    look for a group having ALL the given parameters.

    TODO: This should return a list of groups because the group_name is no longer unique. For now,
    if a user tries to find a group by group_name and there are multiple groups with that name,

    """

    if group_id is None and group_name is None:
        raise GroupInvalidParameters(
            "In order to get group, at least one of group_id, or group_name must be specified"
        )
    query = session.query(Group)
    if group_id is not None:
        query = query.filter(Group.id == group_id)
    if group_name is not None:
        query = query.filter(Group.name == group_name)

    if user_id is not None:
        query = query.join(GroupUser, Group.id == GroupUser.group_id).filter(
            GroupUser.user_id == user_id
        )

    try:
        group = query.one_or_none()
    except MultipleResultsFound:
        raise GroupAlreadyExists(f"Multiple groups exist with name: {group_name}")

    if group is None:
        raise GroupNotFound(f"Did not find group with filters group_id={group_id}")

    return group


def create_group(
    session: Session, group_name: str, user: User, parent_id: Optional[uuid.UUID] = None
) -> Group:
    """
    Creates a new group in the given database session and
    returns the created group object.
    """
    num_user_groups = count_user_groups(session, user.id)
    user_group_limit = get_user_group_limit(session, user.id)
    if num_user_groups >= user_group_limit:
        raise exceptions.UserGroupLimitExceeded(
            f"You have exceeded your quota for group membership ({user_group_limit}). You are in {num_user_groups} groups."
        )

    free_plan_id = get_kv_variable(
        session, "BUGOUT_GROUP_FREE_SUBSCRIPTION_PLAN"
    ).kv_value
    free_plan = subscriptions.get_subscription_plan(session, free_plan_id)

    assert group_name != ""

    # Group name should be stored as lowercase string in the database
    group_name = cast(str, group_name)

    is_child = False
    if parent_id is not None:
        parent_group = session.query(Group).filter(Group.id == parent_id).one_or_none()
        if parent_group is None:
            raise GroupNotFound("There is no provided group with parent id")
        if parent_group.parent is not None:
            raise GroupInvalidParameters(
                "Wrong group parent provided, it is forbidden to have more then one level of inheritance"
            )
        is_child = True

    group = Group(
        name=group_name,
        autogenerated=True if user.autogenerated is True else False,
        parent=parent_id,
    )

    try:
        session.add(group)
        session.commit()
        if not is_child:
            subscriptions.add_group_subscription(
                session,
                group_id=group.id,
                plan_id=free_plan.id,
                units=free_plan.default_units,
                active=True,
            )

        set_user_in_group(
            session=session,
            group_id=group.id,
            current_user_type=Role.owner,
            current_user_autogenerated=True if user.autogenerated is True else False,
            user_type=Role.owner.value,
            username=user.username,
            application_id=user.application_id,
        )
    except Exception as e:
        logger.error(e)
        raise GroupAlreadyExists("Error due adding subscription or set user in group")

    return group


def get_groups_for_user(
    session: Session, user_id: uuid.UUID, include_metrics: bool = True
) -> List[data.GroupUserResponse]:
    """
    Get list of groups current user belongs to.
    """
    groups = (
        session.query(
            Group.id,
            Group.name,
            GroupUser.user_type,
            GroupUser.user_id,
            Group.autogenerated,
            Group.parent,
        )
        .join(Group)
        .filter(user_id == GroupUser.user_id)
        .all()
    )

    groups_response = []

    for group in groups:
        num_users = None
        num_seats = None
        if include_metrics:
            num_users = get_num_users(session, group)
            num_seats = get_num_seats(session, group)

        groups_response.append(
            data.GroupUserResponse(
                group_id=group.id,
                user_id=group.user_id,
                user_type=group.user_type,
                autogenerated=group.autogenerated,
                group_name=group.name,
                parent=group.parent,
                num_users=num_users,
                num_seats=num_seats,
            )
        )

    return groups_response


def count_user_groups(session: Session, user_id: uuid.UUID) -> int:
    """
    Returns the number of groups the given user belongs to.
    """
    query = session.query(func.count(GroupUser.group_id)).filter(
        GroupUser.user_id == user_id
    )
    num_groups = query.one()[0]
    return num_groups


def get_user_group_limit(session: Session, user_id: uuid.UUID) -> int:
    """
    Returns the user's quota for how many groups they can belong to.
    """
    query = session.query(UserGroupLimit).filter(UserGroupLimit.user_id == user_id)
    row = query.one_or_none()
    if row is None:
        return DEFAULT_USER_GROUP_LIMIT
    return row.group_limit


def set_user_group_limit(session: Session, user_id: uuid.UUID, limit: int) -> None:
    """
    Sets the users group limit to the given value.
    """
    query = session.query(UserGroupLimit).filter(UserGroupLimit.user_id == user_id)
    user_group_limit = query.one_or_none()
    if user_group_limit is None:
        user_group_limit = UserGroupLimit(user_id=user_id, group_limit=limit)
        session.add(user_group_limit)
    else:
        user_group_limit.group_limit = limit
    session.commit()


def set_user_in_group(
    session: Session,
    group_id: uuid.UUID,
    current_user_type: Role,
    current_user_autogenerated: bool,
    user_type: str,
    username: Optional[str] = None,
    email: Optional[str] = None,
    application_id: Optional[uuid.UUID] = None,
) -> data.GroupUserResponse:
    """
    Associate user with group or change user type in group if you have permissions.

    This function contains additional arguments to give possibility create groups
    when user doesn't have any roles because group doesn't exist.
    """
    if username is None and email is None:
        raise GroupInvalidParameters(
            "In order to manage user in group, user_type and at least one of username, or email must be specified"
        )

    # Convert to stored format
    if username is not None:
        username = cast(str, username)
        username = username.lower()
        user = get_user(session, username=username, application_id=application_id)
    else:
        user = get_user(session, email=email, application_id=application_id)

    group = get_group(session, group_id=group_id)

    # Check what role proposed user already has
    user_role = (
        session.query(GroupUser)
        .filter(GroupUser.group_id == group.id, GroupUser.user_id == user.id)
        .one_or_none()
    )
    if current_user_autogenerated == False and user.autogenerated == True:
        logger.error("Only autogenerated users allowed to modify autogenerated users")
        raise NoPermissions("You nave no permission to change roles in group")

    # Check group has paren and add user to parent if he doesn't belong to it
    if group.parent is not None:
        parent_user_role = (
            session.query(GroupUser)
            .filter(GroupUser.group_id == group.parent, GroupUser.user_id == user.id)
            .one_or_none()
        )
        if parent_user_role is None:
            raise NoInheritancePermission(
                "User should belong to parent group before he will be added to child group"
            )

    # Check user permissions
    if current_user_type not in Role:
        raise NoPermissions("You nave no permission to change roles in group")

    if current_user_type == Role.member:
        user_type = Role.member.value
    elif current_user_type == Role.admin and user_type == Role.owner.value:
        user_type = Role.admin.value

    role_object = GroupUser(group_id=group.id, user_id=user.id, user_type=user_type)

    # Create new one or update existed
    if user_role is None:
        user_space = get_user_limit(session, group, 1)
        if user_space is False:
            raise LackOfUserSpace("There are no space in group to add new user")

        session.add(role_object)
        session.commit()
    else:
        session.query(GroupUser).filter(
            GroupUser.group_id == group.id, GroupUser.user_id == user.id
        ).update({GroupUser.user_type: user_type})
        session.commit()

    return data.GroupUserResponse(
        group_id=role_object.group_id,
        user_id=role_object.user_id,
        user_type=role_object.user_type,
        group_name=group.name,
        autogenerated=group.autogenerated,
    )


def change_group_name(
    session: Session,
    group_id: uuid.UUID,
    group_name: str,
) -> data.GroupResponse:
    """
    Change group name by group id.
    """
    query = session.query(Group).filter(Group.id == group_id)
    group = query.one_or_none()
    if group is None:
        raise GroupNotFound(f"Did not find any group with group_id={group_id}")

    query.update({Group.name: group_name})
    session.commit()

    return data.GroupResponse(
        id=group.id,
        name=group_name,
        group_name=group_name,
        autogenerated=group.autogenerated,
        subscriptions=[
            subscription_plan.subscription_plan_id
            for subscription_plan in group.subscriptions
        ],
        parent=group.parent,
        created_at=group.created_at,
        updated_at=group.updated_at,
    )


def delete_user_from_group(
    session: Session,
    group_id: uuid.UUID,
    current_user_autogenerated: bool,
    username: Optional[str] = None,
    email: Optional[str] = None,
) -> data.GroupUserResponse:
    """
    Delete user from group.
    """
    if username is None and email is None:
        raise GroupInvalidParameters(
            "In order to add user to group, at least one of username, or email must be specified"
        )

    # Convert to stored format
    if username is not None:
        username = cast(str, username)
        username = username.lower()
        user = get_user(session, username=username)
    else:
        user = get_user(session, email=email)

    group = get_group(session, group_id=group_id)

    # Get role
    role_object = (
        session.query(GroupUser)
        .filter(GroupUser.group_id == group.id, GroupUser.user_id == user.id)
        .one_or_none()
    )
    if role_object is None:
        raise RoleNotFound("Raised when user does not belong group")

    if current_user_autogenerated == False and user.autogenerated == True:
        logger.error("Only autogenerated users allowed to modify autogenerated users")
        raise NoPermissions("You nave no permission to change roles in group")

    session.delete(role_object)
    session.commit()

    # Check group has children and delete user from children groups if he is in list
    children = get_group_children(session, group.id)
    for child_group in children:
        child_role_object = (
            session.query(GroupUser)
            .filter(GroupUser.group_id == child_group.id, GroupUser.user_id == user.id)
            .one_or_none()
        )
        if child_role_object is not None:
            session.delete(child_role_object)
            session.commit()

    return data.GroupUserResponse(
        group_id=role_object.group_id,
        user_id=role_object.user_id,
        user_type=role_object.user_type,
        group_name=group.name,
        autogenerated=group.autogenerated,
    )


def delete_group(session: Session, group_id: uuid.UUID, current_user: User) -> Group:
    """
    Delete a group by group_name. Use get_group to locate the specified group.
    """
    group = get_group(session, group_id=group_id)
    if current_user.autogenerated == False and group.autogenerated == True:
        logger.error("Only autogenerated users allowed to delete autogenerated groups")
        raise NoPermissions("You nave no permission to delete group")
    session.delete(group)
    session.commit()
    return group


def check_user_type_in_group(
    session: Session,
    user_id: uuid.UUID,
    group_id: uuid.UUID,
) -> data.GroupUserResponse:
    """
    Checks if user is an owner or member of requested group.
    """
    group_user = (
        session.query(
            Group.id,
            GroupUser.user_id,
            GroupUser.user_type,
            Group.autogenerated,
            Group.name,
        )
        .join(Group)
        .filter(user_id == GroupUser.user_id)
        .filter(group_id == GroupUser.group_id)
        .one_or_none()
    )

    if group_user is None:
        raise GroupNotFound(f"Did not find available group for user with id={user_id}")

    return data.GroupUserResponse(
        group_id=group_user.id,
        user_id=group_user.user_id,
        user_type=group_user.user_type,
        autogenerated=group_user.autogenerated,
        group_name=group_user.name,
    )


def get_group_users(
    session: Session, group_id: uuid.UUID, group_name: Optional[str]
) -> data.UsersListResponse:
    """
    Extract member list of group.

    # TODO(kompotkot): Optimize and reduce unnecessary SQL queries.
    """
    group = get_group(session, group_id=group_id)

    num_users = get_num_users(session, group)
    max_seats = get_num_seats(session, group)

    group_users_response = data.UsersListResponse(
        id=group_id, name=group_name, users=[], num_users=num_users, num_seats=max_seats
    )

    # Extract users information for requested group
    group_users = (
        session.query(
            GroupUser.group_id,
            Group.name,
            GroupUser.user_id,
            User.username,
            User.email,
            GroupUser.user_type,
        )
        .join(Group, GroupUser.group_id == Group.id)
        .join(User, GroupUser.user_id == User.id)
        .filter(GroupUser.group_id == group_id)
        .all()
    )
    if len(group_users) > 0:
        group_users_response.users = [
            data.UserInListResponse(
                id=user.user_id,
                username=user.username,
                email=user.email,
                user_type=user.user_type,
            )
            for user in group_users
        ]

    return group_users_response


def create_invite(
    db_session: Session,
    group_id: uuid.UUID,
    initiator_user_id: uuid.UUID,
    invited_email: Optional[str] = None,
    user_type: Role = Role.member,
) -> GroupInvite:
    """
    Create invite to group.
    """
    current_time = datetime.utcnow()
    ts_delay = current_time - timedelta(seconds=5)

    invite_object_fresh = (
        db_session.query(GroupInvite)
        .filter(GroupInvite.group_id == group_id, GroupInvite.created_at > ts_delay)
        .one_or_none()
    )
    if invite_object_fresh is not None:
        raise InviteNotAllowed(f"Please wait {ts_delay} until send another invite")

    invite = GroupInvite(
        group_id=group_id,
        initiator_user_id=initiator_user_id,
        invited_email=invited_email,
        user_type=user_type.value,
    )
    db_session.add(invite)
    db_session.commit()

    return invite


def send_group_invite(invite_id: uuid.UUID, email: str, initiator_email: str) -> None:
    """
    Send invite to group for given email.
    """
    group_invite_link = group_invite_link_from_env(str(invite_id), email)
    message = Mail(
        from_email=BUGOUT_FROM_EMAIL,
        to_emails=email,
        subject="Bugout.dev invite to group",
        html_content=(
            f"You have been invited to group by {initiator_email}!\n"
            f"Invite url: {group_invite_link}"
        ),
    )

    try:
        sg = SendGridAPIClient(SENDGRID_API_KEY)
        response = sg.send(message)
    except Exception as e:
        logger.exception(e)
        raise


def list_invites(
    db_session: Session, group_id: uuid.UUID, active: Optional[bool] = None
) -> List[GroupInvite]:
    """
    Return a list of group invites.
    """
    query = db_session.query(GroupInvite).filter(GroupInvite.group_id == group_id)
    if active is not None:
        query = query.filter(GroupInvite.active == active)
    invites = query.all()

    return invites


def get_invite(db_session: Session, invite_id: uuid.UUID) -> GroupInvite:
    """
    Get group invite by id.
    """
    invite = (
        db_session.query(GroupInvite).filter(GroupInvite.id == invite_id).one_or_none()
    )
    if invite is None:
        raise InviteNotFound(f"Invite with id {invite_id} is not presented in database")
    return invite


def update_invite(db_session: Session, invite_id: uuid.UUID, active: bool) -> None:
    """
    Update accepted status of group invite.
    """
    query = db_session.query(GroupInvite).filter(GroupInvite.id == invite_id)
    invite = query.one_or_none()
    if invite is None:
        raise InviteNotFound(f"Invite with id {invite_id} is not presented in database")

    query.update({GroupInvite.active: active})
    db_session.commit()


def get_kv_variable(
    db_session: Session, kv_key: Optional[str] = None, kv_value: Optional[str] = None
) -> KVBrood:
    """
    Extract kv record with key and value from database.
    """
    if kv_key is None and kv_value is None:
        raise KVBroodInvalidParameters(
            "In oder to get kv variables specify at least kv_key or kv_value"
        )

    query = db_session.query(KVBrood)
    if kv_key is not None:
        query = query.filter(KVBrood.kv_key == kv_key)
    if kv_value is not None:
        query = query.filter(KVBrood.kv_value == kv_value)

    kv_variable = query.one_or_none()
    # if kv_variable is None:
    #     raise KVBroodNotFound("Variable not found in KVBrood table")

    return kv_variable


def create_application(
    db_session: Session,
    group_id: uuid.UUID,
    name: str,
    description: Optional[str] = None,
) -> Application:
    """
    Create new record in Application table.
    """
    application = Application(group_id=group_id, name=name, description=description)
    db_session.add(application)
    db_session.commit()

    return application


def get_applications(
    db_session: Session,
    application_id: Optional[uuid.UUID] = None,
    groups_ids: Optional[List[uuid.UUID]] = None,
) -> List[Application]:
    query = db_session.query(Application)

    if application_id is not None:
        query = query.filter(Application.id == application_id)
    if groups_ids is not None:
        query = query.filter(Application.group_id.in_(groups_ids))

    applications = query.all()

    return applications


def delete_application(
    db_session: Session,
    application_id: uuid.UUID,
    groups_ids: List[uuid.UUID],
) -> Application:
    application = (
        db_session.query(Application)
        .filter(Application.id == application_id, Application.group_id.in_(groups_ids))
        .one_or_none()
    )
    if application is None:
        raise exceptions.ApplicationsNotFound(
            f"There are no application with id: {application_id}"
        )

    db_session.delete(application)
    db_session.commit()

    return application



import base64
import json
import logging
from typing import Optional, cast
from uuid import UUID

from fastapi import Depends, HTTPException, Request
from fastapi.exceptions import HTTPException
from fastapi.security.utils import get_authorization_scheme_param
from web3login.auth import to_checksum_address, verify
from web3login.exceptions import (
    Web3AuthorizationExpired,
    Web3AuthorizationWrongApplication,
    Web3VerificationError,
)
from web3login.middlewares.fastapi import OAuth2BearerOrWeb3

from . import actions, data
from .db import yield_db_read_only_session
from .settings import (
    BOT_INSTALLATION_TOKEN,
    BOT_INSTALLATION_TOKEN_HEADER,
    BUGOUT_APPLICATION_ID_HEADER,
)

logger = logging.getLogger(__name__)

# Login implementation follows:
# https://fastapi.tiangolo.com/tutorial/security/simple-oauth2/
oauth2_scheme = OAuth2BearerOrWeb3(tokenUrl="token")
oauth2_scheme_manual = OAuth2BearerOrWeb3(tokenUrl="token", auto_error=False)


async def get_current_user(
    request: Request,
    token: UUID = Depends(oauth2_scheme),
    db_session=Depends(yield_db_read_only_session),
) -> data.UserResponse:
    """
    Middleware returns user if its token or web3 signature verified.
    """
    authorization: str = request.headers.get("Authorization")  # type: ignore
    scheme_raw, _ = get_authorization_scheme_param(authorization)
    scheme = scheme_raw.lower()
    if token is None or token == "":
        raise HTTPException(status_code=404, detail="Access token not found")

    signature_application: str = request.headers.get(BUGOUT_APPLICATION_ID_HEADER)  # type: ignore
    application_id = None
    if signature_application is not None:
        try:
            application_id = cast(UUID, signature_application)
        except Exception:
            raise HTTPException(
                status_code=403, detail="Wrong Web3 signature application provided"
            )

    try:
        if scheme == "web3":
            payload_json = base64.decodebytes(str(token).encode()).decode("utf-8")
            payload = json.loads(payload_json)
            verified = verify(
                authorization_payload=payload,
                application_to_check=str(application_id)
                if application_id is not None
                else "",
            )
            if not verified:
                logger.info("Web3 verification error")
                raise Web3VerificationError()
            web3_address = payload.get("address")
            if web3_address is None:
                logger.error("Web3 address in payload could not be None")
                raise Exception()
            web3_address = to_checksum_address(web3_address)
            user = actions.get_user(
                session=db_session,
                web3_address=web3_address,
                application_id=application_id,
            )

        elif scheme == "bearer":
            is_token_active, user = actions.get_current_user_by_token(
                session=db_session, token=token
            )
            if not is_token_active:
                raise actions.TokenNotActive("Access token not active")
        else:
            raise HTTPException(
                status_code=401, detail="Unaccepted authorization scheme"
            )

    except actions.TokenNotFound as e:
        logger.info(e)
        raise HTTPException(status_code=404, detail="Access token not found")
    except actions.TokenNotActive as e:
        logger.info(e)
        raise HTTPException(status_code=404, detail="Access token not active")
    except actions.UserNotFound as e:
        logger.info(e)
        raise HTTPException(status_code=403, detail="User authorization not found")
    except actions.UserInvalidParameters as e:
        logger.info(e)
        raise HTTPException(status_code=500)
    except Web3AuthorizationExpired:
        raise HTTPException(status_code=403, detail="Signature not verified")
    except Web3AuthorizationWrongApplication:
        raise HTTPException(status_code=403, detail="Signature not verified")
    except Web3VerificationError:
        raise HTTPException(status_code=403, detail="Signature not verified")
    except Exception:
        logger.error("Unhandled exception at get_current_user")
        raise HTTPException(status_code=500)

    return data.UserResponse(
        user_id=user.id,
        username=user.username,
        first_name=user.first_name,
        last_name=user.last_name,
        email=user.email,
        normalized_email=user.normalized_email,
        verified=user.verified,
        created_at=user.created_at,
        updated_at=user.updated_at,
        autogenerated=user.autogenerated,
        application_id=user.application_id,
    )


async def get_current_user_with_groups(
    request: Request,
    token: UUID = Depends(oauth2_scheme),
    db_session=Depends(yield_db_read_only_session),
) -> data.UserWithGroupsResponse:
    """
    Middleware returns user with groups it belongs if its token or web3 signature verified.
    """
    authorization: str = request.headers.get("Authorization")  # type: ignore
    scheme_raw, _ = get_authorization_scheme_param(authorization)
    scheme = scheme_raw.lower()
    if token is None or token == "":
        raise HTTPException(status_code=404, detail="Access token not found")

    signature_application: str = request.headers.get(BUGOUT_APPLICATION_ID_HEADER)  # type: ignore
    application_id = None
    if signature_application is not None:
        try:
            application_id = cast(UUID, signature_application)
        except Exception:
            raise HTTPException(
                status_code=403, detail="Wrong Web3 signature application provided"
            )

    try:
        if scheme == "web3":
            payload_json = base64.decodebytes(str(token).encode()).decode("utf-8")
            payload = json.loads(payload_json)
            verified = verify(
                authorization_payload=payload,
                application_to_check=str(application_id)
                if application_id is not None
                else "",
            )
            if not verified:
                logger.info("Web3 authorization verification error")
                raise Web3VerificationError()
            web3_address = payload.get("address")
            if web3_address is None:
                logger.error("Web3 address in payload could not be None")
                raise Exception()
            web3_address = to_checksum_address(web3_address)
            user_extended = actions.get_user_with_groups(
                session=db_session,
                web3_address=web3_address,
                application_id=application_id,
            )

        elif scheme == "bearer":
            (
                is_token_active,
                user_extended,
            ) = actions.get_current_user_with_groups_by_token(
                session=db_session, token=token
            )
            if not is_token_active:
                raise actions.TokenNotActive("Access token not active")
        else:
            raise HTTPException(
                status_code=401, detail="Unaccepted authorization scheme"
            )

    except actions.TokenNotFound as e:
        logger.info(e)
        raise HTTPException(status_code=404, detail="Access token not found")
    except actions.TokenNotActive as e:
        logger.info(e)
        raise HTTPException(status_code=404, detail="Access token not active")
    except actions.UserNotFound as e:
        logger.info(e)
        raise HTTPException(status_code=403, detail="User authorization not found")
    except actions.UserInvalidParameters as e:
        logger.info(e)
        raise HTTPException(status_code=500)
    except Web3AuthorizationExpired:
        raise HTTPException(status_code=403, detail="Signature not verified")
    except Web3AuthorizationWrongApplication:
        raise HTTPException(status_code=403, detail="Signature not verified")
    except Web3VerificationError:
        raise HTTPException(status_code=403, detail="Signature not verified")
    except Exception:
        logger.error("Unhandled exception at get_current_user_with_groups")
        raise HTTPException(status_code=500)

    return user_extended


def autogenerated_user_token_check(request: Request) -> bool:
    """
    Checks installation token header for autogenerated user access.
    """
    if BOT_INSTALLATION_TOKEN is None or BOT_INSTALLATION_TOKEN == "":
        logger.error("BOT_INSTALLATION_TOKEN environment variable must be set")
        raise HTTPException(status_code=500)

    is_autogenerated_user = False
    installation_token_header: Optional[str] = request.headers.get(
        BOT_INSTALLATION_TOKEN_HEADER, None
    )  # type: ignore
    if (
        installation_token_header is not None
        and BOT_INSTALLATION_TOKEN == installation_token_header
    ):
        is_autogenerated_user = True
    elif (
        installation_token_header is not None
        and BOT_INSTALLATION_TOKEN != installation_token_header
    ):
        raise HTTPException(
            status_code=400,
            detail=f"Invalid {BOT_INSTALLATION_TOKEN_HEADER} provided",
        )
    return is_autogenerated_user


async def is_token_restricted_or_installation(
    request: Request,
    token: UUID = Depends(oauth2_scheme_manual),
    db_session=Depends(yield_db_read_only_session),
) -> bool:
    """
    Allow access if Bugout installation header provided.

    Because of oauth2_scheme_manual we could accept None
    for follow up Bugout header check.
    """
    authorization: str = request.headers.get("Authorization")  # type: ignore
    scheme_raw, _ = get_authorization_scheme_param(authorization)
    scheme = scheme_raw.lower()

    if scheme != "bearer" and scheme is not None and scheme != "":
        raise HTTPException(status_code=400, detail="Unaccepted scheme")

    is_autogenerated_user = autogenerated_user_token_check(request)
    if is_autogenerated_user:
        return False  # Return token.restricted = False
    else:
        token_restricted = await is_token_restricted(request, token, db_session)
        return token_restricted


async def is_token_restricted(
    request: Request,
    token: UUID = Depends(oauth2_scheme),
    db_session=Depends(yield_db_read_only_session),
) -> bool:
    """
    Check if user's token is restricted or not.
    """
    authorization: str = request.headers.get("Authorization")  # type: ignore
    scheme_raw, _ = get_authorization_scheme_param(authorization)
    scheme = scheme_raw.lower()

    if scheme != "bearer":
        raise HTTPException(status_code=400, detail="Unaccepted scheme")

    try:
        token_object = actions.get_token(session=db_session, token=token)
    except actions.TokenNotFound:
        raise HTTPException(status_code=404, detail="Access token not found")
    except Exception:
        logger.error("Unhandled exception at is_token_restricted")
        raise HTTPException(status_code=500)
    return token_object.restricted



"""
Ref to changing an Enum in Python using SQLAlchemy:
https://markrailton.com/blog/creating-migrations-when-changing-an-enum-in-python-using-sql-alchemy
"""
import uuid
from enum import Enum, unique

from sqlalchemy import Boolean, CheckConstraint, Column, DateTime
from sqlalchemy import Enum as PgEnum
from sqlalchemy import (
    ForeignKey,
    Index,
    Integer,
    MetaData,
    PrimaryKeyConstraint,
    String,
)
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql import expression

"""
Naming conventions doc
https://docs.sqlalchemy.org/en/13/core/constraints.html#configuring-constraint-naming-conventions
"""
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=convention)
Base = declarative_base(metadata=metadata)

"""
Creating a utcnow function which runs on the Posgres database server when created_at and updated_at
fields are populated.

Following:
1. https://docs.sqlalchemy.org/en/13/core/compiler.html#utc-timestamp-function
2. https://www.postgresql.org/docs/current/functions-datetime.html#FUNCTIONS-DATETIME-CURRENT
3. https://stackoverflow.com/a/33532154/13659585
"""


class utcnow(expression.FunctionElement):
    type = DateTime


@compiles(utcnow, "postgresql")
def pg_utcnow(element, compiler, **kwargs):
    return "TIMEZONE('utc', statement_timestamp())"


@unique
class Role(Enum):
    """
    Owners able to see payment information and subscribe to plans.
    Admins able to manage group as owners but without access to payment information.
    Members only able to add new members to group without additional permission.
    """

    owner = "owner"
    admin = "admin"
    member = "member"


@unique
class TokenType(Enum):
    bugout = "bugout"
    slack = "slack"
    github = "github"


class User(Base):  # type: ignore
    __tablename__ = "users"

    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
    )
    username = Column(String(100), nullable=False, index=True)
    first_name = Column(String, nullable=True)
    last_name = Column(String, nullable=True)

    # Regular brood user authentication
    email = Column(String, nullable=True)
    normalized_email = Column(String, nullable=True, index=True)
    password_hash = Column(String, nullable=True)

    # Web3 user authentication
    web3_address = Column(String, nullable=True, index=True)

    verified = Column(Boolean, default=False, nullable=False, index=True)
    autogenerated = Column(Boolean, default=False, nullable=False)

    application_id = Column(
        UUID(as_uuid=True),
        ForeignKey(
            "applications.id", name="fk_users_applications_id", ondelete="CASCADE"
        ),
        nullable=True,
    )

    created_at = Column(
        DateTime(timezone=True), server_default=utcnow(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=utcnow(),
        onupdate=utcnow(),
        nullable=False,
    )

    __table_args__ = (
        # Application user uniqueness
        Index(
            "uq_users_username",
            "username",
            unique=True,
            postgresql_where=application_id.is_(None),
        ),
        Index(
            "uq_users_username_application_id",
            "username",
            "application_id",
            unique=True,
            postgresql_where=application_id.isnot(None),
        ),
        Index(
            "uq_users_normalized_email",
            "normalized_email",
            unique=True,
            postgresql_where=application_id.is_(None),
        ),
        Index(
            "uq_users_normalized_email_application_id",
            "normalized_email",
            "application_id",
            unique=True,
            postgresql_where=application_id.isnot(None),
        ),
        Index(
            "uq_users_web3_address",
            "web3_address",
            unique=True,
            postgresql_where=application_id.is_(None),
        ),
        Index(
            "uq_users_web3_address_application_id",
            "web3_address",
            "application_id",
            unique=True,
            postgresql_where=application_id.isnot(None),
        ),
        # Base authentication via email should work together with password
        CheckConstraint(
            "(normalized_email IS NULL) OR (normalized_email IS NOT NULL AND password_hash IS NOT NULL)",
            name="ck_users_normalized_email_password_hash",
        ),
    )

    tokens = relationship(
        "Token", back_populates="user", cascade="all, delete, delete-orphan"
    )
    groups = relationship(
        "GroupUser", back_populates="user", cascade="all, delete, delete-orphan"
    )
    group_limit = relationship(
        "UserGroupLimit", back_populates="user", cascade="all, delete, delete-orphan"
    )


class Token(Base):  # type: ignore
    __tablename__ = "tokens"

    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
        index=True,
    )
    user_id = Column(
        UUID(as_uuid=True),
        ForeignKey("users.id", name="fk_tokens_user_id", ondelete="CASCADE"),
    )
    active = Column(Boolean, default=False, nullable=False, index=True)

    token_type = Column(PgEnum(TokenType, name="token_type"), nullable=False)
    note = Column(String, nullable=True)

    # Restricted tokens cannot perform any operations against the Brood API besides identifying
    # a user
    restricted = Column(Boolean, default=False, nullable=False, index=True)

    created_at = Column(
        DateTime(timezone=True), server_default=utcnow(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=utcnow(),
        onupdate=utcnow(),
        nullable=False,
    )

    user = relationship("User", back_populates="tokens")


class VerificationEmail(Base):  # type: ignore
    __tablename__ = "verification_emails"

    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
    )
    verification_code = Column(String(6))
    user_id = Column(
        UUID(as_uuid=True),
        ForeignKey(
            "users.id", name="fk_verification_emails_user_id", ondelete="CASCADE"
        ),
        index=True,
    )
    active = Column(Boolean, default=False, nullable=False, index=True)
    created_at = Column(
        DateTime(timezone=True), server_default=utcnow(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=utcnow(),
        onupdate=utcnow(),
        nullable=False,
    )


class ResetPassword(Base):  # type: ignore
    __tablename__ = "reset_passwords"

    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
    )
    user_id = Column(
        UUID(as_uuid=True),
        ForeignKey("users.id", name="fk_reset_passwords_user_id", ondelete="CASCADE"),
        index=True,
    )
    completed = Column(Boolean, default=False, nullable=False)
    created_at = Column(
        DateTime(timezone=True), server_default=utcnow(), nullable=False
    )


class Group(Base):  # type: ignore
    __tablename__ = "groups"

    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
    )
    name = Column(String(100), nullable=False)
    parent = Column(
        UUID(as_uuid=True),
        ForeignKey("groups.id", name="fk_groups_parent", ondelete="SET NULL"),
        nullable=True,
    )
    autogenerated = Column(Boolean, default=False, nullable=False)

    created_at = Column(
        DateTime(timezone=True), server_default=utcnow(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=utcnow(),
        onupdate=utcnow(),
        nullable=False,
    )

    user_ids = relationship(
        "GroupUser", back_populates="group", cascade="all, delete, delete-orphan"
    )
    subscriptions = relationship(
        "Subscription", back_populates="group", cascade="all, delete, delete-orphan"
    )


class GroupUser(Base):  # type: ignore
    """
    Relationship between users and groups. User can be
    an owner or a regular member of groups.

    Relevant sqlalchemy documentation on many-many relationships:
    https://docs.sqlalchemy.org/en/13/orm/basic_relationships.html#many-to-many
    """

    __tablename__ = "group_users"
    __table_args__ = (PrimaryKeyConstraint("group_id", "user_id"),)

    group_id = Column(
        UUID(as_uuid=True),
        ForeignKey("groups.id", name="fk_group_users_group_id", ondelete="CASCADE"),
        nullable=False,
    )
    user_id = Column(
        UUID(as_uuid=True),
        ForeignKey("users.id", name="fk_group_users_user_id", ondelete="CASCADE"),
        nullable=False,
    )
    user_type = Column(PgEnum(Role, name="user_type"), nullable=False)

    group = relationship("Group", back_populates="user_ids")
    user = relationship("User", back_populates="groups")


class GroupInvite(Base):  # type: ignore
    """
    Invite to group from user who belongs to this group.
    """

    __tablename__ = "group_invites"
    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
    )
    group_id = Column(
        UUID(as_uuid=True),
        ForeignKey("groups.id", name="fk_group_invites_group_id", ondelete="CASCADE"),
        nullable=False,
    )
    initiator_user_id = Column(
        UUID(as_uuid=True),
        ForeignKey(
            "users.id", name="fk_group_invites_initiator_user_id", ondelete="CASCADE"
        ),
        nullable=False,
    )
    invited_email = Column(String, nullable=True)
    user_type = Column(String, nullable=False)
    active = Column(Boolean, default=True, nullable=False)
    created_at = Column(
        DateTime(timezone=True), server_default=utcnow(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=utcnow(),
        onupdate=utcnow(),
        nullable=False,
    )


class UserGroupLimit(Base):  # type: ignore
    """
    This model is used to impose a limit on how many groups a given user can be a part of (in any
    role).
    """

    __tablename__ = "user_group_limits"

    id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
    )
    user_id = Column(
        UUID(as_uuid=True),
        ForeignKey("users.id", name="fk_user_group_limits_user_id", ondelete="CASCADE"),
        nullable=False,
        unique=True,
    )
    group_limit = Column(Integer, nullable=False)

    user = relationship("User", back_populates="group_limit")


class Subscription(Base):  # type: ignore
    __tablename__ = "subscriptions"
    __table_args__ = (PrimaryKeyConstraint("group_id", "subscription_plan_id"),)

    group_id = Column(
        UUID(as_uuid=True),
        ForeignKey("groups.id", name="fk_subscriptions_group_id", ondelete="CASCADE"),
        nullable=False,
    )
    subscription_plan_id = Column(
        UUID(as_uuid=True),
        ForeignKey(
            "subscription_plans.id",
            name="fk_subscription_plans_subscription_plan_id",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    stripe_customer_id = Column(String, nullable=True)
    stripe_subscription_id = Column(String, nullable=True)
    units = Column(Integer, nullable=False)
    active = Column(Boolean, default=False, nullable=False)

    group = relationship("Group", back_populates="subscriptions")
    plan = relationship("SubscriptionPlan", back_populates="subscriptions")


class SubscriptionPlan(Base):  # type: ignore
    """
    Describes the available subscription plans for organizations.
    """

    __tablename__ = "subscription_plans"

    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
    )
    name = Column(String, nullable=False)
    description = Column(String, nullable=True)
    default_units = Column(Integer, nullable=True)
    plan_type = Column(String, nullable=False)
    public = Column(Boolean, nullable=False)

    stripe_product_id = Column(String, nullable=True)
    stripe_price_id = Column(String, nullable=True)

    subscriptions = relationship(
        "Subscription", back_populates="plan", cascade="all, delete, delete-orphan"
    )


class KVBrood(Base):  # type: ignore
    """
    Special table for unique variables.
    """

    __tablename__ = "kv_brood"

    kv_key = Column(
        String,
        primary_key=True,
        unique=True,
        nullable=False,
    )
    kv_value = Column(
        String,
        nullable=False,
    )


class Application(Base):  # type: ignore
    """
    Resource pooling application.
    """

    __tablename__ = "applications"

    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        default=uuid.uuid4,
        unique=True,
        nullable=False,
    )
    group_id = Column(
        UUID(as_uuid=True),
        ForeignKey("groups.id", name="fk_applications_group_id", ondelete="CASCADE"),
        nullable=False,
    )

    name = Column(String, nullable=False)
    description = Column(String, nullable=True)
